"""
==================================================
Automatic Fiber Bundle Extraction with RecoBundles
==================================================

This example explains how we can use RecoBundles :footcite:p:`Garyfallidis2018`
to extract bundles from tractograms.

First import the necessary modules.
"""

import numpy as np

from dipy.align.streamlinear import whole_brain_slr
from dipy.data import (
    fetch_bundle_atlas_hcp842,
    fetch_target_tractogram_hcp,
    get_bundle_atlas_hcp842,
    get_target_tractogram_hcp,
    get_two_hcp842_bundles,
)
from dipy.io.stateful_tractogram import Space, StatefulTractogram
from dipy.io.streamline import load_trk, save_trk
from dipy.io.utils import create_tractogram_header
from dipy.segment.bundles import RecoBundles
from dipy.viz import actor, window

###############################################################################
# Download and read data for this tutorial

target_file, target_folder = fetch_target_tractogram_hcp()
atlas_file, atlas_folder = fetch_bundle_atlas_hcp842()

atlas_file, all_bundles_files = get_bundle_atlas_hcp842()
target_file = get_target_tractogram_hcp()

sft_atlas = load_trk(atlas_file, "same", bbox_valid_check=False)
atlas = sft_atlas.streamlines
atlas_header = create_tractogram_header(atlas_file, *sft_atlas.space_attributes)

sft_target = load_trk(target_file, "same", bbox_valid_check=False)
target = sft_target.streamlines
target_header = create_tractogram_header(target_file, *sft_target.space_attributes)

###############################################################################
# let's visualize atlas tractogram and target tractogram before registration

interactive = False

scene = window.Scene()
scene.SetBackground(1, 1, 1)
scene.add(actor.line(atlas, colors=(1, 0, 1)))
scene.add(actor.line(target, colors=(1, 1, 0)))
window.record(scene=scene, out_path="tractograms_initial.png", size=(600, 600))
if interactive:
    window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Atlas and target before registration.
#
#
# We will register target tractogram to model atlas' space using streamlinear
# registration (SLR) :footcite:p:`Garyfallidis2015`.

moved, transform, qb_centroids1, qb_centroids2 = whole_brain_slr(
    atlas,
    target,
    x0="affine",
    verbose=True,
    progressive=True,
    rng=np.random.default_rng(1984),
)


###############################################################################
# We save the transform generated in this registration, so that we can use
# it in the bundle profiles example

np.save("slr_transform.npy", transform)

###############################################################################
# let's visualize atlas tractogram and target tractogram after registration

interactive = False

scene = window.Scene()
scene.SetBackground(1, 1, 1)
scene.add(actor.line(atlas, colors=(1, 0, 1)))
scene.add(actor.line(moved, colors=(1, 1, 0)))
window.record(
    scene=scene, out_path="tractograms_after_registration.png", size=(600, 600)
)
if interactive:
    window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Atlas and target after registration.
#
#
#
# Extracting bundles using RecoBundles :footcite:p:`Garyfallidis2018`
#
# RecoBundles requires a model (reference) bundle and tries to extract similar
# looking bundle from the input tractogram. There are some key parameters that
# users can set as per requirements. Here are the key threshold parameters
# measured in millimeters and their function in Recobundles:
#
# - model_clust_thr : It will use QuickBundles to get the centroids of the
#   model bundle and work with centroids instead of all streamlines. This
#   helps to make RecoBundles faster. The larger the value of the threshold,
#   the fewer centroids will be, the and smaller the threshold value, the
#   more centroids will be. If you prefer to use all the streamlines of the
#   model bundle, you can set this threshold to 0.01 mm.
#   Recommended range of the model_clust_thr is 0.01 - 3.0 mm.
#
# - reduction_thr : This threshold will be used to reduce the search space
#   for finding the streamlines that match model bundle streamlines in shape.
#   Instead of looking at the entire tractogram, now we will be looking at
#   neighboring region of a model bundle in the tractogram. Increase the
#   threshold to increase the search space.
#   Recommended range of the reduction_thr is 15 - 30 mm.
#
# - pruning_thr : This threshold will filter the streamlines for which the
#   distance to the model bundle is greater than the pruning_thr.
#   This serves to filter the neighborhood area (search space) to get
#   streamlines that are like the model bundle.
#   Recommended range of the pruning_thr is 8 - 12 mm.
#
# - reduction_distance and pruning_distance : Distance method used
#   internally. Minimum Diferect Flip distance (mdf) or Mean Average Minimum
#   (mam). Default is set to mdf.
#
# - slr : If slr flag is set to True, local registration of model bundle
#   with neighbouring area will be performed. Default and recommended is True.
#
#
#
# Read Arcuate Fasciculus Left and Corticospinal Tract Left bundles from
# already fetched atlas data to use them as model bundle. Let's visualize the
# Arcuate Fasciculus Left model bundle.

model_af_l_file, model_cst_l_file = get_two_hcp842_bundles()
sft_af_l = load_trk(model_af_l_file, reference="same", bbox_valid_check=False)
model_af_l = sft_af_l.streamlines

interactive = False

scene = window.Scene()
scene.SetBackground(1, 1, 1)
scene.add(actor.line(model_af_l))
scene.set_camera(
    focal_point=(-18.17281532, -19.55606842, 6.92485857),
    position=(-360.11, -30.46, -40.44),
    view_up=(-0.03, 0.028, 0.89),
)
window.record(scene=scene, out_path="AF_L_model_bundle.png", size=(600, 600))
if interactive:
    window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Model Arcuate Fasciculus Left bundle

rb = RecoBundles(moved, verbose=True, rng=np.random.default_rng(2001))

recognized_af_l, af_l_labels = rb.recognize(
    model_bundle=model_af_l,
    model_clust_thr=0.1,
    reduction_thr=15,
    pruning_thr=7,
    reduction_distance="mdf",
    pruning_distance="mdf",
    slr=True,
)

###############################################################################
# let's visualize extracted Arcuate Fasciculus Left bundle

interactive = False

scene = window.Scene()
scene.SetBackground(1, 1, 1)
scene.add(actor.line(recognized_af_l.copy()))
scene.set_camera(
    focal_point=(-18.17281532, -19.55606842, 6.92485857),
    position=(-360.11, -30.46, -40.44),
    view_up=(-0.03, 0.028, 0.89),
)
window.record(scene=scene, out_path="AF_L_recognized_bundle.png", size=(600, 600))
if interactive:
    window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Extracted Arcuate Fasciculus Left bundle
#
#
#
# Save the bundle as a trk file. Let's save the recognized bundle in the
# common space (atlas space), in this case, MNI space.

reco_af_l = StatefulTractogram(recognized_af_l, atlas_header, Space.RASMM)
save_trk(reco_af_l, "AF_L_rec_1.trk", bbox_valid_check=False)

###############################################################################
# Let's save the recognized bundle in the original space of the subject
# anatomy.

reco_af_l = StatefulTractogram(target[af_l_labels], target_header, Space.RASMM)
save_trk(reco_af_l, "AF_L_org_1.trk", bbox_valid_check=False)

###############################################################################
# Now, let's increase the reduction_thr and pruning_thr values.

recognized_af_l, af_l_labels = rb.recognize(
    model_bundle=model_af_l,
    model_clust_thr=0.1,
    reduction_thr=20,
    pruning_thr=10,
    reduction_distance="mdf",
    pruning_distance="mdf",
    slr=True,
)

###############################################################################
# let's visualize extracted Arcuate Fasciculus Left bundle.

interactive = False

scene = window.Scene()
scene.SetBackground(1, 1, 1)
scene.add(actor.line(recognized_af_l.copy()))
scene.set_camera(
    focal_point=(-18.17281532, -19.55606842, 6.92485857),
    position=(-360.11, -30.46, -40.44),
    view_up=(-0.03, 0.028, 0.89),
)
window.record(scene=scene, out_path="AF_L_recognized_bundle2.png", size=(600, 600))
if interactive:
    window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Extracted Arcuate Fasciculus Left bundle
#
#
# Save the bundle as a trk file. Let's save the recognized bundle in the
# common space (atlas space), in this case, MNI space.

reco_af_l = StatefulTractogram(recognized_af_l, atlas_header, Space.RASMM)
save_trk(reco_af_l, "AF_L_rec_2.trk", bbox_valid_check=False)

###############################################################################
# Let's save the recognized bundle in the original space of the subject
# anatomy.

reco_af_l = StatefulTractogram(target[af_l_labels], target_header, Space.RASMM)
save_trk(reco_af_l, "AF_L_org_2.trk", bbox_valid_check=False)

###############################################################################
# Now, let's increase the reduction_thr and pruning_thr values further.

recognized_af_l, af_l_labels = rb.recognize(
    model_bundle=model_af_l,
    model_clust_thr=0.1,
    reduction_thr=25,
    pruning_thr=12,
    reduction_distance="mdf",
    pruning_distance="mdf",
    slr=True,
)

###############################################################################
# let's visualize extracted Arcuate Fasciculus Left bundle.

interactive = False

scene = window.Scene()
scene.SetBackground(1, 1, 1)
scene.add(actor.line(recognized_af_l.copy()))
scene.set_camera(
    focal_point=(-18.17281532, -19.55606842, 6.92485857),
    position=(-360.11, -30.46, -40.44),
    view_up=(-0.03, 0.028, 0.89),
)
window.record(scene=scene, out_path="AF_L_recognized_bundle3.png", size=(600, 600))
if interactive:
    window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Extracted Arcuate Fasciculus Left bundle
#
#
# Save the bundle as a trk file. Let's save the recognized bundle in the
# common space (atlas space), in this case, MNI space.

reco_af_l = StatefulTractogram(recognized_af_l, atlas_header, Space.RASMM)
save_trk(reco_af_l, "AF_L_rec_3.trk", bbox_valid_check=False)

###############################################################################
# Let's save the recognized bundle in the original space of the subject
# anatomy.

reco_af_l = StatefulTractogram(target[af_l_labels], target_header, Space.RASMM)
save_trk(reco_af_l, "AF_L_org_3.trk", bbox_valid_check=False)


###############################################################################
# Let's apply auto-calibrated RecoBundles on the output of standard
# RecoBundles. This step will filter out the outlier streamlines. This time,
# the RecoBundles' extracted bundle will serve as a model bundle. As a rule of
# thumb, provide larger threshold values in standard RecoBundles function and
# smaller values in the auto-calibrated RecoBundles (refinement) step.

r_recognized_af_l, r_af_l_labels = rb.refine(
    model_bundle=model_af_l,
    pruned_streamlines=recognized_af_l,
    model_clust_thr=0.1,
    reduction_thr=15,
    pruning_thr=6,
    reduction_distance="mdf",
    pruning_distance="mdf",
    slr=True,
)

###############################################################################
# let's visualize extracted refined Arcuate Fasciculus Left bundle.

interactive = False

scene = window.Scene()
scene.SetBackground(1, 1, 1)
scene.add(actor.line(r_recognized_af_l.copy()))
scene.set_camera(
    focal_point=(-18.17281532, -19.55606842, 6.92485857),
    position=(-360.11, -30.46, -40.44),
    view_up=(-0.03, 0.028, 0.89),
)
window.record(
    scene=scene, out_path="AF_L_refine_recognized_bundle.png", size=(600, 600)
)
if interactive:
    window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Extracted Arcuate Fasciculus Left bundle
#
#
#
# Save the bundle as a trk file. Let's save the recognized bundle in the
# common space (atlas space), in this case, MNI space.

reco_af_l = StatefulTractogram(r_recognized_af_l, atlas_header, Space.RASMM)
save_trk(reco_af_l, "AF_L_rec_refine.trk", bbox_valid_check=False)

###############################################################################
# Let's save the recognized bundle in the original space of the subject
# anatomy.

reco_af_l = StatefulTractogram(target[r_af_l_labels], target_header, Space.RASMM)
save_trk(reco_af_l, "AF_L_org_refine.trk", bbox_valid_check=False)

###############################################################################
# Let's load Corticospinal Tract Left model bundle and visualize it.

sft_cst_l = load_trk(model_cst_l_file, "same", bbox_valid_check=False)
model_cst_l = sft_cst_l.streamlines

interactive = False

scene = window.Scene()
scene.SetBackground(1, 1, 1)
scene.add(actor.line(model_cst_l))
scene.set_camera(
    focal_point=(-18.17281532, -19.55606842, 6.92485857),
    position=(-360.11, -30.46, -40.44),
    view_up=(-0.03, 0.028, 0.89),
)
window.record(scene=scene, out_path="CST_L_model_bundle.png", size=(600, 600))
if interactive:
    window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# The Corticospinal tract model bundle

recognized_cst_l, cst_l_labels = rb.recognize(
    model_bundle=model_cst_l,
    model_clust_thr=0.1,
    reduction_thr=15,
    pruning_thr=7,
    reduction_distance="mdf",
    pruning_distance="mdf",
    slr=True,
)

###############################################################################
# let's visualize extracted Corticospinal tract Left bundle

interactive = False

scene = window.Scene()
scene.SetBackground(1, 1, 1)
scene.add(actor.line(recognized_cst_l.copy()))
scene.set_camera(
    focal_point=(-18.17281532, -19.55606842, 6.92485857),
    position=(-360.11, -30.46, -40.44),
    view_up=(-0.03, 0.028, 0.89),
)
window.record(scene=scene, out_path="CST_L_recognized_bundle.png", size=(600, 600))
if interactive:
    window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Extracted Corticospinal tract Left bundle
#
#
#
# Save the bundle as a trk file. Let's save the recognized bundle in the
# common space (atlas space), in this case, MNI space.

reco_cst_l = StatefulTractogram(recognized_cst_l, atlas_header, Space.RASMM)
save_trk(reco_cst_l, "CST_L_rec_1.trk", bbox_valid_check=False)

###############################################################################
# Let's save the recognized bundle in the original space of the subject
# anatomy.

reco_cst_l = StatefulTractogram(target[cst_l_labels], target_header, Space.RASMM)
save_trk(reco_cst_l, "CST_L_org_1.trk", bbox_valid_check=False)

###############################################################################
# Now, let's increase the reduction_thr and pruning_thr values.

recognized_cst_l, cst_l_labels = rb.recognize(
    model_bundle=model_cst_l,
    model_clust_thr=0.1,
    reduction_thr=20,
    pruning_thr=10,
    reduction_distance="mdf",
    pruning_distance="mdf",
    slr=True,
)

###############################################################################
# let's visualize extracted Corticospinal tract Left bundle.

interactive = False

scene = window.Scene()
scene.SetBackground(1, 1, 1)
scene.add(actor.line(recognized_cst_l.copy()))
scene.set_camera(
    focal_point=(-18.17281532, -19.55606842, 6.92485857),
    position=(-360.11, -30.46, -40.44),
    view_up=(-0.03, 0.028, 0.89),
)
window.record(scene=scene, out_path="CST_L_recognized_bundle2.png", size=(600, 600))
if interactive:
    window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Extracted Corticospinal tract Left bundle
#
#
# Save the bundle as a trk file. Let's save the recognized bundle in the
# common space (atlas space), in this case, MNI space.

reco_cst_l = StatefulTractogram(recognized_cst_l, atlas_header, Space.RASMM)
save_trk(reco_cst_l, "CST_L_rec_2.trk", bbox_valid_check=False)

###############################################################################
# Let's save the recognized bundle in the original space of the subject
# anatomy.

reco_cst_l = StatefulTractogram(target[cst_l_labels], target_header, Space.RASMM)
save_trk(reco_cst_l, "CST_L_org_2.trk", bbox_valid_check=False)

###############################################################################
# Now, let's increase the reduction_thr and pruning_thr values further.

recognized_cst_l, cst_l_labels = rb.recognize(
    model_bundle=model_cst_l,
    model_clust_thr=0.1,
    reduction_thr=25,
    pruning_thr=12,
    reduction_distance="mdf",
    pruning_distance="mdf",
    slr=True,
)

###############################################################################
# let's visualize extracted Corticospinal tract Left bundle.

interactive = False

scene = window.Scene()
scene.SetBackground(1, 1, 1)
scene.add(actor.line(recognized_cst_l.copy()))
scene.set_camera(
    focal_point=(-18.17281532, -19.55606842, 6.92485857),
    position=(-360.11, -30.46, -40.44),
    view_up=(-0.03, 0.028, 0.89),
)
window.record(scene=scene, out_path="CST_L_recognized_bundle3.png", size=(600, 600))
if interactive:
    window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Extracted Corticospinal tract Left bundle
#
#
#
#
# Save the bundle as a trk file. Let's save the recognized bundle in the
# common space (atlas space), in this case, MNI space.

reco_cst_l = StatefulTractogram(recognized_cst_l, atlas_header, Space.RASMM)
save_trk(reco_cst_l, "CST_L_rec_3.trk", bbox_valid_check=False)

###############################################################################
# Let's save the recognized bundle in the original space of the subject
# anatomy.

reco_cst_l = StatefulTractogram(target[cst_l_labels], target_header, Space.RASMM)
save_trk(reco_cst_l, "CST_L_org_3.trk", bbox_valid_check=False)

###############################################################################
# Let's apply auto-calibrated RecoBundles on the output of standard
# RecoBundles. This step will filter out the outlier streamlines. This time,
# the RecoBundles' extracted bundle will serve as a model bundle. As a rule of
# thumb, provide larger threshold values in standard RecoBundles function and
# smaller values in the auto-calibrated RecoBundles (refinement) step.

r_recognized_cst_l, r_cst_l_labels = rb.refine(
    model_bundle=model_cst_l,
    pruned_streamlines=recognized_cst_l,
    model_clust_thr=0.1,
    reduction_thr=15,
    pruning_thr=6,
    reduction_distance="mdf",
    pruning_distance="mdf",
    slr=True,
)

###############################################################################
# let's visualize extracted refined Corticospinal tract Left bundle.

interactive = False

scene = window.Scene()
scene.SetBackground(1, 1, 1)
scene.add(actor.line(r_recognized_cst_l.copy()))
scene.set_camera(
    focal_point=(-18.17281532, -19.55606842, 6.92485857),
    position=(-360.11, -30.46, -40.44),
    view_up=(-0.03, 0.028, 0.89),
)
window.record(
    scene=scene, out_path="CST_L_refine_recognized_bundle.png", size=(600, 600)
)
if interactive:
    window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Extracted refined Corticospinal tract Left bundle
#
#
# Save the bundle as a trk file. Let's save the recognized bundle in the
# common space (atlas space), in this case, MNI space.

reco_cst_l = StatefulTractogram(r_recognized_cst_l, atlas_header, Space.RASMM)
save_trk(reco_cst_l, "CST_L_rec_refine.trk", bbox_valid_check=False)

###############################################################################
# Let's save the recognized bundle in the original space of the subject
# anatomy.

reco_cst_l = StatefulTractogram(target[r_cst_l_labels], target_header, Space.RASMM)
save_trk(reco_cst_l, "CST_L_org_refine.trk", bbox_valid_check=False)

###############################################################################
# This example shows how changing different threshold parameters can change the
# output. It is up to the user to decide what is the desired output and select
# parameters accordingly.
#
# References
# ----------
#
# .. footbibliography::
#



"""
==========================================================
Signal Reconstruction Using Spherical Harmonics
==========================================================

This example shows how you can use a spherical harmonics (SH) function to
reconstruct any spherical function using DIPY_. In order to generate a
signal, we will need to generate an evenly distributed sphere.

Let's import some standard packages.
"""

import numpy as np

from dipy.core.sphere import HemiSphere, Sphere, disperse_charges
from dipy.data import get_sphere
from dipy.reconst.shm import sf_to_sh, sh_to_sf
from dipy.sims.voxel import multi_tensor_odf
from dipy.viz import actor, window

###############################################################################
# We can first create some random points on a ``HemiSphere`` using spherical
# polar coordinates.

rng = np.random.default_rng()
n_pts = 64
theta = np.pi * rng.random(n_pts)
phi = 2 * np.pi * rng.random(n_pts)
hsph_initial = HemiSphere(theta=theta, phi=phi)

###############################################################################
# Next, we call ``disperse_charges`` which will iteratively move the points so
# that the electrostatic potential energy is minimized. In ``hsph_updated`` we
# have the updated ``HemiSphere`` with the points nicely distributed on the
# hemisphere.

hsph_updated, potential = disperse_charges(hsph_initial, 5000)
sphere = Sphere(xyz=np.vstack((hsph_updated.vertices, -hsph_updated.vertices)))

###############################################################################
# We now need to create our initial signal. To do so, we will use our sphere's
# vertices as the sampled points of our spherical function (SF). We will
# use ``multi_tensor_odf`` to simulate an ODF. For more information on how to
# use DIPY_ to simulate a signal and ODF, see
# :ref:`sphx_glr_examples_built_simulations_simulate_multi_tensor.py`.

mevals = np.array([[0.0015, 0.00015, 0.00015], [0.0015, 0.00015, 0.00015]])
angles = [(0, 0), (60, 0)]
odf = multi_tensor_odf(sphere.vertices, mevals, angles, [50, 50])

# Enables/disables interactive visualization
interactive = False

scene = window.Scene()
scene.SetBackground(1, 1, 1)

odf_actor = actor.odf_slicer(odf[None, None, None, :], sphere=sphere)
odf_actor.RotateX(90)
scene.add(odf_actor)

print("Saving illustration as symm_signal.png")
window.record(scene=scene, out_path="symm_signal.png", size=(300, 300))
if interactive:
    window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Illustration of the simulated signal sampled on a sphere of 64 points
# per hemisphere
#
#
# We can now express this signal as a series of SH coefficients using
# ``sf_to_sh``. This function converts a series of SF coefficients in a series
# of SH coefficients. For more information on SH basis, see :ref:`sh-basis`.
# For this example, we will use the ``descoteaux07`` basis up to a maximum SH
# order of 8.

# Change this value to try out other bases
sh_basis = "descoteaux07"
# Change this value to try other maximum orders
sh_order_max = 8

sh_coeffs = sf_to_sh(odf, sphere, sh_order_max=sh_order_max, basis_type=sh_basis)

###############################################################################
# ``sh_coeffs`` is an array containing the SH coefficients multiplying the SH
# functions of the chosen basis. We can use it as input of ``sh_to_sf`` to
# reconstruct our original signal. We will now reproject our signal on a high
# resolution sphere using ``sh_to_sf``.

high_res_sph = get_sphere(name="symmetric724").subdivide(n=2)
reconst = sh_to_sf(
    sh_coeffs, high_res_sph, sh_order_max=sh_order_max, basis_type=sh_basis
)

scene.clear()
odf_actor = actor.odf_slicer(reconst[None, None, None, :], sphere=high_res_sph)
odf_actor.RotateX(90)
scene.add(odf_actor)

print("Saving output as symm_reconst.png")
window.record(scene=scene, out_path="symm_reconst.png", size=(300, 300))
if interactive:
    window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Reconstruction of a symmetric signal on a high resolution sphere using a
# symmetric basis
#
#
# While a symmetric SH basis works well for reconstructing symmetric SF, it
# fails to do so on asymmetric signals. We will now create such a signal by
# using a different ODF for each hemisphere of our sphere.

mevals = np.array([[0.0015, 0.0003, 0.0003]])
angles = [(0, 0)]
odf2 = multi_tensor_odf(sphere.vertices, mevals, angles, [100])

n_pts_hemisphere = int(sphere.vertices.shape[0] / 2)
asym_odf = np.append(odf[:n_pts_hemisphere], odf2[n_pts_hemisphere:])

scene.clear()
odf_actor = actor.odf_slicer(asym_odf[None, None, None, :], sphere=sphere)
odf_actor.RotateX(90)
scene.add(odf_actor)

print("Saving output as asym_signal.png")
window.record(scene=scene, out_path="asym_signal.png", size=(300, 300))
if interactive:
    window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Illustration of an asymmetric signal sampled on a sphere of 64
# points per hemisphere
#
#
# Let's try to reconstruct this SF using a symmetric SH basis.

sh_coeffs = sf_to_sh(asym_odf, sphere, sh_order_max=sh_order_max, basis_type=sh_basis)
reconst = sh_to_sf(
    sh_coeffs, high_res_sph, sh_order_max=sh_order_max, basis_type=sh_basis
)

scene.clear()
odf_actor = actor.odf_slicer(reconst[None, None, None, :], sphere=high_res_sph)
odf_actor.RotateX(90)
scene.add(odf_actor)

print("Saving output as asym_reconst.png")
window.record(scene=scene, out_path="asym_reconst.png", size=(300, 300))
if interactive:
    window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Reconstruction of an asymmetric signal using a symmetric SH basis
#
#
# As we can see, a symmetric basis fails to properly represent asymmetric SF.
# Fortunately, DIPY_ also implements full SH bases, which can deal with
# symmetric as well as asymmetric signals. For this tutorial, we will
# demonstrate it using the ``descoteaux07`` full SH basis by setting
# ``full_basis=true``.

sh_coeffs = sf_to_sh(
    asym_odf, sphere, sh_order_max=sh_order_max, basis_type=sh_basis, full_basis=True
)
reconst = sh_to_sf(
    sh_coeffs,
    high_res_sph,
    sh_order_max=sh_order_max,
    basis_type=sh_basis,
    full_basis=True,
)

scene.clear()
odf_actor = actor.odf_slicer(reconst[None, None, None, :], sphere=high_res_sph)
odf_actor.RotateX(90)
scene.add(odf_actor)

print("Saving output as asym_reconst_full.png")
window.record(scene=scene, out_path="asym_reconst_full.png", size=(300, 300))
if interactive:
    window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Reconstruction of an asymmetric signal using a full SH basis
#
#
# As we can see, a full SH basis properly reconstruct asymmetric signal.



"""Attempt to generate templates for module reference with Sphinx.

To include extension modules, first identify them as valid in the
``_uri2path`` method, then handle them in the ``_parse_module_with_import``
script.

Notes
-----
This parsing is based on the import and introspection of modules.
Previously functions and classes were found by parsing the text of .py files.

Extension modules should be discovered and included as well.

This is a modified version of a script originally shipped with the PyMVPA
project, then adapted for use first in NIPY and then in skimage. PyMVPA
is an MIT-licensed project.

"""

import ast
from importlib import import_module
from inspect import getmodule, ismethod
import os
import re
from types import BuiltinFunctionType, FunctionType

# suppress print statements (warnings for empty files)
DEBUG = True


class ApiDocWriter:
    """Automatic detection and parsing of API docs to Sphinx-parsable reST format."""

    # only separating first two levels
    rst_section_levels = ["*", "=", "-", "~", "^"]

    def __init__(
        self,
        package_name,
        rst_extension=".rst",
        package_skip_patterns=None,
        module_skip_patterns=None,
        object_skip_patterns=None,
        other_defines=True,
    ):
        r"""Initialize package for parsing.

        Parameters
        ----------
        package_name : string
            Name of the top-level package.  *package_name* must be the
            name of an importable package
        rst_extension : string, optional
            Extension for reST files, default '.rst'
        package_skip_patterns : None or sequence of {strings, regexps}
            Sequence of strings giving URIs of packages to be excluded
            Operates on the package path, starting at (including) the
            first dot in the package path, after *package_name* - so,
            if *package_name* is ``sphinx``, then ``sphinx.util`` will
            result in ``.util`` being passed for searching by these
            regexps.  If is None, gives default. Default is:
            ['\.tests$']
        module_skip_patterns : None or sequence
            Sequence of strings giving URIs of modules to be excluded
            Operates on the module name including preceding URI path,
            back to the first dot after *package_name*.  For example
            ``sphinx.util.console`` results in the string to search of
            ``.util.console``
            If is None, gives default. Default is:
            ['\.setup$', '\._']
        object_skip_patterns: None or sequence of {strings, regexps}
            skip some specific class or function
        other_defines : {True, False}, optional
            Whether to include classes and functions that are imported in a
            particular module but not defined there.

        """
        if package_skip_patterns is None:
            package_skip_patterns = ["\\.tests$"]
        if module_skip_patterns is None:
            module_skip_patterns = ["\\.setup$", "\\._"]
        if object_skip_patterns is None:
            object_skip_patterns = []

        self.root_path = ""
        self.written_modules = None
        self._package_name = None
        self.package_name = package_name
        self.rst_extension = rst_extension
        self.package_skip_patterns = package_skip_patterns
        self.module_skip_patterns = module_skip_patterns
        self.object_skip_patterns = object_skip_patterns
        self.other_defines = other_defines

    def get_package_name(self):
        return self._package_name

    def set_package_name(self, package_name):
        """Set package_name.

        >>> docwriter = ApiDocWriter('sphinx')
        >>> import sphinx
        >>> docwriter.root_path == sphinx.__path__[0]
        True
        >>> docwriter.package_name = 'docutils'
        >>> import docutils
        >>> docwriter.root_path == docutils.__path__[0]
        True

        """
        # It's also possible to imagine caching the module parsing here
        self._package_name = package_name
        root_module = self._import(package_name)
        if root_module.__path__:
            self.root_path = root_module.__path__[-1]

        if not self.root_path or 'editable_loader' in self.root_path.lower():
            self.root_path = os.path.dirname(root_module.__file__)
        self.written_modules = None

    package_name = property(
        get_package_name, set_package_name, None, "get/set package_name"
    )

    @staticmethod
    def _import(name):
        """Import namespace package."""
        mod = __import__(name)
        components = name.split(".")
        for comp in components[1:]:
            mod = getattr(mod, comp)
        return mod

    @staticmethod
    def _get_object_name(line):
        """Get second token in line.

        Examples
        --------
        >>> docwriter = ApiDocWriter('sphinx')
        >>> docwriter._get_object_name("  def func():  ")
        'func'
        >>> docwriter._get_object_name("  class Klass:  ")
        'Klass'
        >>> docwriter._get_object_name("  class Klass:  ")
        'Klass'

        """
        name = line.split()[1].split("(")[0].strip()
        # in case we have classes which are not derived from object
        # ie. old style classes
        return name.rstrip(":")

    def _uri2path(self, uri):
        """Convert uri to absolute filepath.

        Parameters
        ----------
        uri : string
            URI of python module to return path for

        Returns
        -------
        path : None or string
            Returns None if there is no valid path for this URI
            Otherwise returns absolute file system path for URI

        Examples
        --------
        >>> docwriter = ApiDocWriter('sphinx')
        >>> import sphinx
        >>> modpath = sphinx.__path__[0]
        >>> res = docwriter._uri2path('sphinx.builder')
        >>> res == os.path.join(modpath, 'builder.py')
        True
        >>> res = docwriter._uri2path('sphinx')
        >>> res == os.path.join(modpath, '__init__.py')
        True
        >>> docwriter._uri2path('sphinx.does_not_exist')

        """
        if uri == self.package_name:
            return os.path.join(self.root_path, "__init__.py")
        path = uri.replace(self.package_name + ".", "")
        path = path.replace(".", os.path.sep)
        path = os.path.join(self.root_path, path)
        # XXX maybe check for extensions as well?
        if os.path.exists(path + ".py"):  # file
            path += ".py"
        elif os.path.exists(path + ".pyx"):  # file
            path += ".pyx"
        elif os.path.exists(os.path.join(path, "__init__.py")):
            path = os.path.join(path, "__init__.py")
        else:
            return None
        return path

    def _path2uri(self, dirpath):
        """Convert directory path to uri."""
        package_dir = self.package_name.replace(".", os.path.sep)
        relpath = dirpath.replace(self.root_path, package_dir)
        if relpath.startswith(os.path.sep):
            relpath = relpath[1:]
        return relpath.replace(os.path.sep, ".")

    def _parse_module(self, uri):
        """Parse module defined in *uri*."""
        filename = self._uri2path(uri)
        if filename is None:
            print(filename, "erk")
            # nothing that we could handle here.
            return ([], [])

        f = open(filename, "rt")
        functions, classes = self._parse_lines(f)
        f.close()
        return functions, classes

    def _parse_module_with_import(self, uri):
        """Look for functions and classes in an importable module.

        Parameters
        ----------
        uri : str
            The name of the module to be parsed. This module needs to be
            importable.

        Returns
        -------
        functions : list of str
            A list of (public) function names in the module.
        classes : list of str
            A list of (public) class names in the module.
        constants : list of str
            A list of (public) constant names in the module.

        """
        mod = import_module(uri)
        patterns = f"(?:{'|'.join(self.object_skip_patterns)})"
        pat = re.compile(patterns)

        if mod.__file__.endswith('.py'):
            with open(mod.__file__, encoding="utf8") as fi:
                node = ast.parse(fi.read())

            functions = []
            classes = []
            constants = []
            for n in node.body:
                if not hasattr(n, "name"):
                    if not isinstance(n, ast.Assign):
                        continue

                if isinstance(n, ast.ClassDef):
                    if n.name.startswith("_") or pat.search(n.name):
                        continue
                    classes.append(n.name)
                elif isinstance(n, ast.FunctionDef):
                    if n.name.startswith("_") or pat.search(n.name):
                        continue
                    functions.append(n.name)
                # Specific condition for vtk and fury
                elif isinstance(n, ast.Assign):
                    try:
                        if isinstance(n.value, ast.Call):
                            if hasattr(n.targets[0], "id"):
                                func_name = n.targets[0].id
                            else:
                                continue

                            if func_name.startswith("_") or pat.search(func_name) or  \
                               isinstance(n.targets[0], ast.Tuple):
                                continue
                            if isinstance(n.value, (ast.Constant, ast.Dict, ast.Tuple)):
                                constants.append(n.targets[0].id)
                                continue
                            functions.append(n.targets[0].id)
                        elif hasattr(n.value, "attr"):
                            if n.value.attr.startswith("vtk"):
                                classes.append(n.targets[0].id)
                    except Exception as e:
                        print(mod.__file__)
                        print(n.lineno)
                        print(n.targets[0])
                        print(str(e))

            return functions, classes, constants
        else:
            # find all public objects in the module.
            obj_strs = [obj for obj in dir(mod) if not obj.startswith("_")]
            functions = []
            classes = []
            constants = []
            for obj_str in obj_strs:
                # find the actual object from its string representation
                if obj_str not in mod.__dict__:
                    continue
                obj = mod.__dict__[obj_str]
                # Check if function / class defined in module
                if not self.other_defines and not getmodule(obj) == mod:
                    continue
                # figure out if obj is a function or class
                if (hasattr(obj, "func_name") or isinstance(obj, BuiltinFunctionType) or
                   ismethod(obj) or isinstance(obj, FunctionType)):
                    functions.append(obj_str)
                else:
                    try:
                        issubclass(obj, object)
                        classes.append(obj_str)
                    except TypeError:
                        # not a function or class
                        pass
            return functions, classes, constants

    def _parse_lines(self, linesource):
        """Parse lines of text for functions and classes."""
        functions = []
        classes = []
        for line in linesource:
            if line.startswith("def ") and line.count("("):
                # exclude private stuff
                name = self._get_object_name(line)
                if not name.startswith("_"):
                    functions.append(name)
            elif line.startswith("class "):
                # exclude private stuff
                name = self._get_object_name(line)
                if not name.startswith("_"):
                    classes.append(name)
            else:
                pass
        functions.sort()
        classes.sort()
        return functions, classes

    def generate_api_doc(self, uri):
        """Make autodoc documentation template string for a module.

        Parameters
        ----------
        uri : string
            python location of module - e.g 'sphinx.builder'

        Returns
        -------
        head : string
            Module name, table of contents.
        body : string
            Function and class docstrings.

        """
        # get the names of all classes and functions
        functions, classes, constants = self._parse_module_with_import(uri)
        if not len(functions) and not len(classes) and DEBUG:
            print("WARNING: Empty -", uri)  # dbg

        # Make a shorter version of the uri that omits the package name for
        # titles
        uri_short = re.sub(r"^%s\." % self.package_name, "", uri)

        head = ".. AUTO-GENERATED FILE -- DO NOT EDIT!\n\n"
        body = ""

        # Set the chapter title to read 'module' for all modules except for the
        # main packages
        if "." in uri_short:
            title = "Module: :mod:`" + uri_short + "`"
            head += title + "\n" + self.rst_section_levels[2] * len(title)
        else:
            title = ":mod:`" + uri_short + "`"
            head += title + "\n" + self.rst_section_levels[1] * len(title)

        head += "\n.. automodule:: " + uri + "\n"

        head += "\n.. currentmodule:: " + uri + "\n"

        body += "\n.. currentmodule:: " + uri + "\n\n"

        for c in constants:
            # must NOT exclude from index to keep cross-refs working
            body += c + "\n"
            body += self.rst_section_levels[3] * len(c) + "\n"
            body += "\n.. autodata:: " + c + "\n"
            body += "   :no-value:\n"
            body += "   :annotation:\n"

            body += "\n\n"

        for c in classes:
            body += f"\n:class:`{c}`\n{self.rst_section_levels[3] * (len(c) + 9)}\n\n"
            body += "\n.. autoclass:: " + c + "\n"
            # must NOT exclude from index to keep cross-refs working
            body += "  :members:\n" "  :undoc-members:\n" "  :show-inheritance:\n" "\n"
        head += ".. autosummary::\n\n"
        for f in constants + classes + functions:
            head += "   " + f + "\n"
        head += "\n\n"

        for f in functions:
            # must NOT exclude from index to keep cross-refs working
            body += f + "\n"
            body += self.rst_section_levels[3] * len(f) + "\n"
            body += "\n.. autofunction:: " + f + "\n"
            body += "\n\n"

        return head, body

    def _survives_exclude(self, matchstr, match_type):
        r"""Return True if *matchstr* does not match patterns.

        ``self.package_name`` removed from front of string if present

        Examples
        --------
        >>> dw = ApiDocWriter('sphinx')
        >>> dw._survives_exclude('sphinx.okpkg', 'package')
        True
        >>> dw.package_skip_patterns.append('^\\.badpkg$')
        >>> dw._survives_exclude('sphinx.badpkg', 'package')
        False
        >>> dw._survives_exclude('sphinx.badpkg', 'module')
        True
        >>> dw._survives_exclude('sphinx.badmod', 'module')
        True
        >>> dw.module_skip_patterns.append('^\\.badmod$')
        >>> dw._survives_exclude('sphinx.badmod', 'module')
        False

        """
        if match_type == "module":
            patterns = self.module_skip_patterns
        elif match_type == "package":
            patterns = self.package_skip_patterns
        else:
            raise ValueError(f"Cannot interpret match type '{match_type}'")
        # Match to URI without package name
        L = len(self.package_name)
        if matchstr[:L] == self.package_name:
            matchstr = matchstr[L:]
        for pat in patterns:
            try:
                _ = pat.search
            except AttributeError:
                pat = re.compile(pat)
            if pat.search(matchstr):
                return False

        return True

    def discover_modules(self):
        r"""Return module sequence discovered from ``self.package_name``.

        Parameters
        ----------
        None

        Returns
        -------
        mods : sequence
            Sequence of module names within ``self.package_name``

        Examples
        --------
        >>> dw = ApiDocWriter('sphinx')
        >>> mods = dw.discover_modules()
        >>> 'sphinx.util' in mods
        True
        >>> dw.package_skip_patterns.append('\.util$')
        >>> 'sphinx.util' in dw.discover_modules()
        False

        """
        modules = [self.package_name]
        # raw directory parsing
        for dirpath, dirnames, filenames in os.walk(self.root_path):
            # Check directory names for packages
            root_uri = self._path2uri(os.path.join(self.root_path, dirpath))

            # Normally, we'd only iterate over dirnames, but since
            # dipy does not import a whole bunch of modules we'll
            # include those here as well (the *.py filenames).
            filenames = [os.path.splitext(f)[0] for f in filenames
                         if (f.endswith(".py") and not f.startswith("__init__")) or
                         f.endswith(".pyx")]

            for subpkg_name in dirnames + filenames:
                package_uri = ".".join((root_uri, subpkg_name))
                package_path = self._uri2path(package_uri)
                if package_path and self._survives_exclude(package_uri, "package"):
                    modules.append(package_uri)

        return sorted(modules)

    def write_modules_api(self, modules, outdir):
        # upper-level modules
        ulms = [
            ".".join(m.split(".")[:2]) if m.count(".") >= 1 else m.split(".")[0]
            for m in modules
        ]

        from collections import OrderedDict

        module_by_ulm = OrderedDict()

        for v, k in zip(modules, ulms):
            if k in module_by_ulm:
                module_by_ulm[k].append(v)
            else:
                module_by_ulm[k] = [v]

        written_modules = []

        for ulm, mods in module_by_ulm.items():
            print(f"Generating docs for {ulm}:")
            document_head = []
            document_body = []

            for m in mods:
                print(f"  -> {m}")
                head, body = self.generate_api_doc(m)

                document_head.append(head)
                document_body.append(body)

            out_module = ulm + self.rst_extension
            outfile = os.path.join(outdir, out_module)
            fileobj = open(outfile, "wt")

            fileobj.writelines(document_head + document_body)
            fileobj.close()
            written_modules.append(out_module)

        self.written_modules = written_modules

    def write_api_docs(self, outdir):
        """Generate API reST files.

        Parameters
        ----------
        outdir : string
            Directory name in which to store files
            We create automatic filenames for each module

        Returns
        -------
        None

        Notes
        -----
        Sets self.written_modules to list of written modules

        """
        if not os.path.exists(outdir):
            os.mkdir(outdir)
        # compose list of modules
        modules = self.discover_modules()
        self.write_modules_api(modules, outdir)

    def write_index(self, outdir, froot="gen", relative_to=None):
        """Make a reST API index file from written files.

        Parameters
        ----------
        path : string
            Filename to write an index to
        outdir : string
            Directory to which to write generated index file
        froot : string, optional
            root (filename without extension) of filename to write to
            Defaults to 'gen'.  We add ``self.rst_extension``.
        relative_to : string
            path to which written filenames are relative.  This
            component of the written file path will be removed from
            outdir, in the generated index.  Default is None, meaning,
            leave the path as it is.

        """
        if self.written_modules is None:
            raise ValueError("No modules written")
        # Get full filename path
        path = os.path.join(outdir, froot + self.rst_extension)
        # Path written into index is relative to rootpath
        if relative_to is not None:
            relpath = (outdir + os.path.sep).replace(relative_to + os.path.sep, "")
        else:
            relpath = outdir
        idx = open(path, "wt")
        w = idx.write
        w(".. AUTO-GENERATED FILE -- DO NOT EDIT!\n\n")

        title = "API Reference"
        w(title + "\n")
        w("=" * len(title) + "\n\n")
        w(".. toctree::\n\n")
        for f in self.written_modules:
            w("   %s\n" % os.path.join(relpath, f))
        idx.close()



"""A unified interface for performing and debugging optimization problems."""

import abc
import warnings

import numpy as np
import scipy.optimize as opt
from scipy.optimize import minimize

from dipy.testing.decorators import warning_for_keywords
from dipy.utils.optpkg import optional_package

cvxpy, have_cvxpy, _ = optional_package("cvxpy", min_version="1.4.1")


class Optimizer:
    @warning_for_keywords()
    def __init__(
        self,
        fun,
        x0,
        args=(),
        *,
        method="L-BFGS-B",
        jac=None,
        hess=None,
        hessp=None,
        bounds=None,
        constraints=(),
        tol=None,
        callback=None,
        options=None,
        evolution=False,
    ):
        """A class for handling minimization of scalar function of one or more
        variables.

        Parameters
        ----------
        fun : callable
            Objective function.

        x0 : ndarray
            Initial guess.

        args : tuple, optional
            Extra arguments passed to the objective function and its
            derivatives (Jacobian, Hessian).

        method : str, optional
            Type of solver.  Should be one of

                - 'Nelder-Mead'
                - 'Powell'
                - 'CG'
                - 'BFGS'
                - 'Newton-CG'
                - 'Anneal'
                - 'L-BFGS-B'
                - 'TNC'
                - 'COBYLA'
                - 'SLSQP'
                - 'dogleg'
                - 'trust-ncg'

        jac : bool or callable, optional
            Jacobian of objective function. Only for CG, BFGS, Newton-CG,
            dogleg, trust-ncg.
            If `jac` is a Boolean and is True, `fun` is assumed to return the
            value of Jacobian along with the objective function. If False, the
            Jacobian will be estimated numerically.
            `jac` can also be a callable returning the Jacobian of the
            objective. In this case, it must accept the same arguments
            as `fun`.

        hess, hessp : callable, optional
            Hessian of objective function or Hessian of objective function
            times an arbitrary vector p.  Only for Newton-CG,
            dogleg, trust-ncg.
            Only one of `hessp` or `hess` needs to be given.  If `hess` is
            provided, then `hessp` will be ignored.  If neither `hess` nor
            `hessp` is provided, then the hessian product will be approximated
            using finite differences on `jac`. `hessp` must compute the Hessian
            times an arbitrary vector.

        bounds : sequence, optional
            Bounds for variables (only for L-BFGS-B, TNC and SLSQP).
            ``(min, max)`` pairs for each element in ``x``, defining
            the bounds on that parameter. Use None for one of ``min`` or
            ``max`` when there is no bound in that direction.

        constraints : dict or sequence of dict, optional
            Constraints definition (only for COBYLA and SLSQP).
            Each constraint is defined in a dictionary with fields:

                type : str
                    Constraint type: 'eq' for equality, 'ineq' for inequality.
                fun : callable
                    The function defining the constraint.
                jac : callable, optional
                    The Jacobian of `fun` (only for SLSQP).
                args : sequence, optional
                    Extra arguments to be passed to the function and Jacobian.

            Equality constraint means that the constraint function result is to
            be zero whereas inequality means that it is to be non-negative.
            Note that COBYLA only supports inequality constraints.

        tol : float, optional
            Tolerance for termination. For detailed control, use
            solver-specific options.

        callback : callable, optional
            Called after each iteration, as ``callback(xk)``, where ``xk`` is
            the current parameter vector. Only available using Scipy >= 0.12.

        options : dict, optional
            A dictionary of solver options. All methods accept the following
            generic options:

                maxiter : int
                    Maximum number of iterations to perform.
                disp : bool
                    Set to True to print convergence messages.

            For method-specific options, see
            `show_options('minimize', method)`.

        evolution : bool, optional
            save history of x for each iteration. Only available using Scipy
            >= 0.12.

        See Also
        --------
        scipy.optimize.minimize

        """
        self.size_of_x = len(x0)
        self._evol_kx = None

        if evolution is True:
            self._evol_kx = []

            def history_of_x(kx):
                self._evol_kx.append(kx)

            res = minimize(
                fun,
                x0,
                args,
                method,
                jac,
                hess,
                hessp,
                bounds,
                constraints,
                tol,
                callback=history_of_x,
                options=options,
            )

        else:
            res = minimize(
                fun,
                x0,
                args,
                method,
                jac,
                hess,
                hessp,
                bounds,
                constraints,
                tol,
                callback,
                options,
            )

        self.res = res

    @property
    def xopt(self):
        return self.res["x"]

    @property
    def fopt(self):
        return self.res["fun"]

    @property
    def nit(self):
        return self.res["nit"]

    @property
    def nfev(self):
        return self.res["nfev"]

    @property
    def message(self):
        return self.res["message"]

    def print_summary(self):
        print(self.res)

    @property
    def evolution(self):
        if self._evol_kx is not None:
            return np.asarray(self._evol_kx)
        else:
            return None


def spdot(A, B):
    """The same as np.dot(A, B), except it works even if A or B or both
    are sparse matrices.

    Parameters
    ----------
    A, B : arrays of shape (m, n), (n, k)

    Returns
    -------
    The matrix product AB. If both A and B are sparse, the result will be a
    sparse matrix. Otherwise, a dense result is returned

    See discussion here:
    http://mail.scipy.org/pipermail/scipy-user/2010-November/027700.html

    """
    return A @ B


@warning_for_keywords()
def sparse_nnls(
    y,
    X,
    *,
    momentum=1,
    step_size=0.01,
    non_neg=True,
    check_error_iter=10,
    max_error_checks=10,
    converge_on_sse=0.99,
):
    """
    Solve y=Xh for h, using gradient descent, with X a sparse matrix.

    Parameters
    ----------
    y : 1-d array of shape (N)
        The data. Needs to be dense.

    X : ndarray. May be either sparse or dense. Shape (N, M)
       The regressors

    momentum : float, optional
        The persistence of the gradient.

    step_size : float, optional
        The increment of parameter update in each iteration

    non_neg : Boolean, optional
        Whether to enforce non-negativity of the solution.

    check_error_iter : int, optional
        How many rounds to run between error evaluation for
        convergence-checking.

    max_error_checks : int, optional
        Don't check errors more than this number of times if no improvement in
        r-squared is seen.

    converge_on_sse : float, optional
      a percentage improvement in SSE that is required each time to say
      that things are still going well.

    Returns
    -------
    h_best : The best estimate of the parameters.

    """
    num_regressors = X.shape[1]
    # Initialize the parameters at the origin:
    h = np.zeros(num_regressors)
    # If nothing good happens, we'll return that:
    h_best = h
    iteration = 1
    ss_residuals_min = np.inf  # This will keep track of the best solution
    sse_best = np.inf  # This will keep track of the best performance so far
    count_bad = 0  # Number of times estimation error has gone up.
    error_checks = 0  # How many error checks have we done so far

    while 1:
        if iteration > 1:
            # The gradient is (Kay 2008 supplemental page 27):
            gradient = spdot(X.T, spdot(X, h) - y)
            gradient += momentum * gradient
            # Normalize to unit-length
            unit_length_gradient = gradient / np.sqrt(np.dot(gradient, gradient))
            # Update the parameters in the direction of the gradient:
            h -= step_size * unit_length_gradient
            if non_neg:
                # Set negative values to 0:
                h[h < 0] = 0

        # Every once in a while check whether it's converged:
        if np.mod(iteration, check_error_iter):
            # This calculates the sum of squared residuals at this point:
            sse = np.sum((y - spdot(X, h)) ** 2)
            # Did we do better this time around?
            if sse < ss_residuals_min:
                # Update your expectations about the minimum error:
                ss_residuals_min = sse
                h_best = h  # This holds the best params we have so far
                # Are we generally (over iterations) converging on
                # sufficient improvement in r-squared?
                if sse < converge_on_sse * sse_best:
                    sse_best = sse
                    count_bad = 0
                else:
                    count_bad += 1
            else:
                count_bad += 1

            if count_bad >= max_error_checks:
                return h_best
            error_checks += 1
        iteration += 1


class SKLearnLinearSolver(metaclass=abc.ABCMeta):
    """
    Provide a sklearn-like uniform interface to algorithms that solve problems
    of the form: $y = Ax$ for $x$

    Sub-classes of SKLearnLinearSolver should provide a 'fit' method that have
    the following signature: `SKLearnLinearSolver.fit(X, y)`, which would set
    an attribute `SKLearnLinearSolver.coef_`, with the shape (X.shape[1],),
    such that an estimate of y can be calculated as:
    `y_hat = np.dot(X, SKLearnLinearSolver.coef_.T)`
    """

    def __init__(self, *args, **kwargs):
        self._args = args
        self._kwargs = kwargs

    @abc.abstractmethod
    def fit(self, X, y):
        """Implement for all derived classes"""

    def predict(self, X):
        """
        Predict using the result of the model

        Parameters
        ----------
        X : array-like (n_samples, n_features)
            Samples.

        Returns
        -------
        C : array, shape = (n_samples,)
            Predicted values.
        """
        X = np.asarray(X)
        return np.dot(X, self.coef_.T)


class NonNegativeLeastSquares(SKLearnLinearSolver):
    """
    A sklearn-like interface to scipy.optimize.nnls

    """

    def fit(self, X, y):
        """
        Fit the NonNegativeLeastSquares linear model to data

        Parameters
        ----------
        X : array-like (n_samples, n_features)
            Samples.
        y : array-like (n_samples,)
            Target values.

        """
        coef, rnorm = opt.nnls(X, y)
        self.coef_ = coef
        return self


class PositiveDefiniteLeastSquares:
    @warning_for_keywords()
    def __init__(self, m, *, A=None, L=None):
        r"""Regularized least squares with linear matrix inequality constraints.

        See :footcite:p:`DelaHaije2020` for further details about the method.

        Generate a CVXPY representation of a regularized least squares
        optimization problem subject to linear matrix inequality constraints.

        Parameters
        ----------
        m : int
            Positive int indicating the number of regressors.
        A : array (t = m + k + 1, p, p), optional
            Constraint matrices $A$.
        L : array (m, m), optional
            Regularization matrix $L$.
            Default: None.

        Notes
        -----
        The basic problem is to solve for $h$ the minimization of

        $c=\|X h - y\|^2 + \|L h\|^2$,

        where $X$ is an (m, m) upper triangular design matrix and $y$ is a set
        of m measurements, subject to the constraint that

        $M=A_0+\sum_{i=0}^{m-1} h_i A_{i+1}+\sum_{j=0}^{k-1} s_j A_{m+j+1}>0$,

        where $s_j$ are slack variables and where the inequality sign denotes
        positive definiteness of the matrix $M$. The sparsity pattern and size
        of $X$ and $y$ are fixed, because every design matrix and set of
        measurements can be reduced to an equivalent (minimal) formulation of
        this type.

        This formulation is used here mainly to enforce polynomial
        sum-of-squares constraints on various models, as described in
        :footcite:p:`DelaHaije2020`.

        References
        ----------
        .. footbibliography::
        """
        # Note: several comments refer to DelaHaije2020 eq 22

        # Input
        self.A = A  # list: [zeros, H(omega), L(alpha)] --- see eq 22
        self.L = L  # this is a regularizer matrix, NOT L(alpha) eq 22

        # Problem size
        t = len(A) if A else 0
        k = t - m - 1  # length of alpha in L(alpha) eq 22

        # Unknowns
        self._X = cvxpy.Parameter((m, m))  # Design matrix
        self._f = cvxpy.Parameter(m)  # Given solution for feasibility check
        self._h = cvxpy.Variable(m)  # Solution to constrained problem
        self._y = cvxpy.Parameter(m)  # Regressand

        # Error output
        self._zeros = np.zeros(m)

        # Objective
        c = self._X @ self._h - self._y
        if L is not None:
            c += L @ self._h

        f_objective = cvxpy.Minimize(0)
        p_objective = cvxpy.Minimize(cvxpy.norm(c))

        # Constraints
        if t:
            M = F = A[0]  # first matrix all zeros (use to initialize)
            if k > 0:
                for i in range(m):  # loop over H(omega) from eq 22
                    F += self._f[i] * A[i + 1]
                    M += self._h[i] * A[i + 1]
                    # A-matrix for 'intercept' (22nd, i=21) should be zeros
                self._s = cvxpy.Variable(k)
                for j in range(k):  # loop over L(alpha) from eq 22
                    F += self._s[j] * A[m + j + 1]
                    M += self._s[j] * A[m + j + 1]
            else:
                for i in range(t - 1):
                    F += self._f[i] * A[i + 1]
                    M += self._h[i] * A[i + 1]
            f_constraints = [F >> 0]
            p_constraints = [M >> 0]
        else:
            f_constraints = p_constraints = []

        # CVXPY problems
        self.problem = cvxpy.Problem(p_objective, p_constraints)
        self.unconstrained_problem = cvxpy.Problem(p_objective)
        self.feasibility_problem = cvxpy.Problem(f_objective, f_constraints)

    @warning_for_keywords()
    def solve(self, design_matrix, measurements, *, check=False, **kwargs):
        r"""Solve CVXPY problem

        Solve a CVXPY problem instance for a given design matrix and a given set
        of observations, and return the optimum.

        Parameters
        ----------
        design_matrix : array (n, m)
            Design matrix.
        measurements : array (n)
            Measurements.
        check : boolean, optional
            If True check whether the unconstrained optimization solution
            already satisfies the constraints, before running the constrained
            optimization. This adds overhead, but can avoid unnecessary
            constrained optimization calls.
        kwargs : keyword arguments
            Arguments passed to the CVXPY solve method.

        Returns
        -------
        h : array (m)
             Estimated optimum for problem variables $h$.
        """

        # Compute and set reduced problem parameters
        try:
            X = np.linalg.cholesky(np.dot(design_matrix.T, design_matrix)).T
        except np.linalg.linalg.LinAlgError:
            msg = "Cholesky decomposition failed, returning zero array. Verify "
            msg += "that the data is sufficient to estimate the model "
            msg += "parameters, and that the design matrix has full rank."
            warnings.warn(msg, stacklevel=2)
            return self._zeros
        self._X.value = X
        self._y.value = np.linalg.multi_dot(
            [X, np.linalg.pinv(design_matrix), measurements]
        )

        try:
            # Check unconstrained solution
            if check:
                # Solve unconstrained problem
                self.unconstrained_problem.solve(**kwargs)

                # Return zeros if optimization failed
                status = self.unconstrained_problem.status
                if status != "optimal":
                    msg = f"Solver failed to produce an optimum: {status}."
                    warnings.warn(msg, stacklevel=2)
                    msg = "Optimization failed, returning zero array."
                    warnings.warn(msg, stacklevel=2)
                    return self._zeros

                # Return unconstrained solution if satisfactory
                self._f.value = self._h.value
                self.feasibility_problem.solve(**kwargs)
                if self.feasibility_problem.status == "optimal":
                    return np.asarray(self._h.value).squeeze()

            # Solve constrained problem
            self.problem.solve(**kwargs)

            # Show warning if solution is not optimal
            status = self.problem.status
            if status != "optimal":
                msg = f"Solver failed to produce an optimum: {status}."
                warnings.warn(msg, stacklevel=2)

            # Return solution
            return np.asarray(self._h.value).squeeze()

        except cvxpy.error.SolverError:
            # Return zeros
            msg = "Optimization failed, returning zero array."
            warnings.warn(msg, stacklevel=2)
            return self._zeros



"""
==================================================================
Continuous and analytical diffusion signal modelling with 3D-SHORE
==================================================================

We show how to model the diffusion signal as a linear combination
of continuous functions from the SHORE basis :footcite:p:`Merlet2013`,
:footcite:p:`Ozarslan2008`, :footcite:p:`Ozarslan2009`. We also compute the
analytical Orientation Distribution Function (ODF).

First import the necessary modules:
"""

from dipy.core.gradients import gradient_table
from dipy.data import get_fnames, get_sphere
from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti
from dipy.reconst.shore import ShoreModel
from dipy.viz import actor, window

###############################################################################
# Download and read the data for this tutorial.

# ``fetch_isbi2013_2shell()`` provides data from the `ISBI HARDI contest 2013
# <http://hardi.epfl.ch/static/events/2013_ISBI/>`_ acquired for two shells at
# b-values 1500 $s/mm^2$ and 2500 $s/mm^2$.

# The six parameters of these two functions define the ROI where to reconstruct
# the data. They respectively correspond to ``(xmin,xmax,ymin,ymax,zmin,zmax)``
# with x, y, z and the three axis defining the spatial positions of the voxels.

fraw, fbval, fbvec = get_fnames(name="isbi2013_2shell")

data, affine = load_nifti(fraw)
bvals, bvecs = read_bvals_bvecs(fbval, fbvec)
gtab = gradient_table(bvals, bvecs=bvecs)
data_small = data[10:40, 22, 10:40]

print(f"data.shape {data.shape}")

###############################################################################
# ``data`` contains the voxel data and ``gtab`` contains a ``GradientTable``
# object (gradient information e.g. b-values). For example, to show the
# b-values it is possible to write::
#
#     print(gtab.bvals)
#
# Instantiate the SHORE Model.
#
# ``radial_order`` is the radial order of the SHORE basis.
#
# ``zeta`` is the scale factor of the SHORE basis.
#
# ``lambdaN`` and ``lambdaL`` are the radial and angular regularization
# constants, respectively.
#
# For details regarding these four parameters see :footcite:p:`Cheng2011` and
# :footcite:p:`Merlet2013`.

radial_order = 6
zeta = 700
lambdaN = 1e-8
lambdaL = 1e-8
asm = ShoreModel(
    gtab, radial_order=radial_order, zeta=zeta, lambdaN=lambdaN, lambdaL=lambdaL
)

###############################################################################
# Fit the SHORE model to the data

asmfit = asm.fit(data_small)

###############################################################################
# Load an odf reconstruction sphere

sphere = get_sphere(name="repulsion724")

###############################################################################
# Compute the ODFs

odf = asmfit.odf(sphere)
print(f"odf.shape {odf.shape}")

###############################################################################
# Display the ODFs

# Enables/disables interactive visualization
interactive = False

scene = window.Scene()
sfu = actor.odf_slicer(odf[:, None, :], sphere=sphere, colormap="plasma", scale=0.5)
sfu.RotateX(-90)
sfu.display(y=0)
scene.add(sfu)
window.record(scene=scene, out_path="odfs.png", size=(600, 600))
if interactive:
    window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Orientation distribution functions.
#
#
# References
# ----------
#
# .. footbibliography::
#



# Init file for visualization package
import warnings

from dipy.utils.optpkg import optional_package
from dipy.viz.horizon.app import horizon

fury_pckg_msg = (
    "You do not have FURY installed. Some visualization functions might not "
    "work for you. Please install or upgrade FURY using pip install -U fury. "
    "For detailed installation instructions visit: https://fury.gl/"
)

# Allow import, but disable doctests if we don't have fury
fury, has_fury, _ = optional_package(
    "fury", trip_msg=fury_pckg_msg, min_version="0.10.0"
)


if has_fury:
    from fury import actor, colormap, interactor, lib, shaders, ui, utils, window
    from fury.data import DATA_DIR as FURY_DATA_DIR, fetch_viz_icons, read_viz_icons

else:
    warnings.warn(fury_pckg_msg, stacklevel=2)

# We make the visualization requirements optional imports:
_, has_mpl, _ = optional_package(
    "matplotlib",
    trip_msg="You do not have Matplotlib installed. Some visualization "
    "functions might not work for you. For installation instructions, "
    "please visit: https://matplotlib.org/",
)

if has_mpl:
    from . import projections





cimport numpy as cnp
import numpy as np
import numpy.testing as npt

from dipy.core.math cimport f_array_min, f_array_max, f_max, f_min


def test_f_array_max():
    cdef int size = 10
    cdef double[:] d_arr = np.arange(size, dtype=np.float64)
    cdef float[:] f_arr = np.arange(size, dtype=np.float32)

    npt.assert_equal(f_array_max(&d_arr[0], size), 9)
    npt.assert_equal(f_array_max(&f_arr[0], size), 9)


def test_f_array_min():
    cdef int size = 6
    cdef double[:] d_arr = np.arange(5, 5+size, dtype=np.float64)
    cdef float[:] f_arr = np.arange(5, 5+size, dtype=np.float32)

    npt.assert_almost_equal(f_array_min(&d_arr[0], size), 5)
    npt.assert_almost_equal(f_array_min(&f_arr[0], size), 5)


def test_f_max():
    npt.assert_equal(f_max[double](1, 2), 2)
    npt.assert_equal(f_max[float](2, 1), 2)
    npt.assert_equal(f_max(1., 1.), 1)


def test_f_min():
    npt.assert_equal(f_min[double](1, 2), 1)
    npt.assert_equal(f_min[float](2, 1), 1)
    npt.assert_equal(f_min(1.0, 1.0), 1)




"""
====================================
Parallel reconstruction using Q-Ball
====================================

We show an example of parallel reconstruction using a Q-Ball Constant Solid
Angle model (see Aganj et al. (MRM 2010)) and `peaks_from_model`.

Import modules, fetch and read data, and compute the mask.
"""

import time

from dipy.core.gradients import gradient_table
from dipy.data import get_fnames, get_sphere
from dipy.direction import peaks_from_model
from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti
from dipy.reconst.shm import CsaOdfModel
from dipy.segment.mask import median_otsu

hardi_fname, hardi_bval_fname, hardi_bvec_fname = get_fnames(name="stanford_hardi")

data, affine = load_nifti(hardi_fname)

bvals, bvecs = read_bvals_bvecs(hardi_bval_fname, hardi_bvec_fname)
gtab = gradient_table(bvals, bvecs=bvecs)

maskdata, mask = median_otsu(
    data, vol_idx=range(10, 50), median_radius=3, numpass=1, autocrop=True, dilate=2
)

###############################################################################
# We instantiate our CSA model with spherical harmonic order ($l$) of 4

csamodel = CsaOdfModel(gtab, 4)

###############################################################################
# `Peaks_from_model` is used to calculate properties of the ODFs (Orientation
# Distribution Function) and return for
# example the peaks and their indices, or GFA which is similar to FA but for
# ODF based models. This function mainly needs a reconstruction model, the
# data and a sphere as input. The sphere is an object that represents the
# spherical discrete grid where the ODF values will be evaluated.

sphere = get_sphere(name="repulsion724")

start_time = time.time()

###############################################################################
# We will first run `peaks_from_model` using parallelism with 2 processes. If
# `num_processes` is None (default option) then this function will find the
# total number of processors from the operating system and use this number as
# `num_processes`. Sometimes it makes sense to use only a few of the processes
# in order to allow resources for other applications. However, most of the
# times using the default option will be sufficient.

csapeaks_parallel = peaks_from_model(
    model=csamodel,
    data=maskdata,
    sphere=sphere,
    relative_peak_threshold=0.5,
    min_separation_angle=25,
    mask=mask,
    return_odf=False,
    normalize_peaks=True,
    npeaks=5,
    parallel=True,
    num_processes=2,
)

time_parallel = time.time() - start_time
print(f"peaks_from_model using 2 processes ran in : {time_parallel} seconds")

###############################################################################
# If we don't use parallelism then we need to set `parallel=False`:

start_time = time.time()
csapeaks = peaks_from_model(
    model=csamodel,
    data=maskdata,
    sphere=sphere,
    relative_peak_threshold=0.5,
    min_separation_angle=25,
    mask=mask,
    return_odf=False,
    normalize_peaks=True,
    npeaks=5,
    parallel=False,
    num_processes=None,
)

time_single = time.time() - start_time
print(f"peaks_from_model ran in : {time_single} seconds")

print(f"Speedup factor : {time_single / time_parallel}")

###############################################################################
# In Windows if you get a runtime error about frozen executable please start
# your script by adding your code above in a ``main`` function and use::
#
#    if __name__ == '__main__':
#        import multiprocessing
#        multiprocessing.freeze_support()
#        main()



from dataclasses import dataclass, field
import fnmatch
import os
from os.path import join as pjoin
from pathlib import Path
import shutil
import sys

from sphinx.util import logging

try:
    import tomllib
except ImportError:
    import tomli as tomllib

logger = logging.getLogger(__name__)

sphx_glr_sep = '########################################'
sphx_glr_sep += '#######################################'
sphx_glr_sep_2 = '#%%'


@dataclass(order=True)
class examplesConfig:
    sort_index: int = field(init=False)
    readme: str
    position: int
    enable: bool
    files: list
    folder_name: str

    def __post_init__(self):
        self.sort_index = self.position


def abort(error):
    print(f'*WARNING* Examples Revamp not generated: \n\n{error}')
    exit()


def already_converted(content):
    """Check if the example file is already converted to sphinx-gallery format.

    Parameters
    ----------
    content : list of str
        example file content

    Returns
    -------
    bool
        _description_
    """
    return content.count(sphx_glr_sep) >= 2 or \
        content.count(sphx_glr_sep_2) >= 2


def convert_to_sphinx_gallery_format(content):
    """Convert the example file to sphinx-gallery format.

    Parameters
    ----------
    content : list of str
        example file content

    Returns
    -------
    list of str
        example file content converted to sphinx-gallery format
    """
    inheader = True
    indocs = False

    new_content = ''
    for line in content:

        if inheader:
            if not indocs and (line.startswith('"""') or line.startswith("'''") or
               line.startswith('r"""') or line.startswith("r'''")):
                new_content += line
                if len(line.rstrip()) < 5:
                    indocs = True
                else:
                    # single line doc
                    inheader = False
                continue

            if line.rstrip().endswith('"""') or line.rstrip().endswith("'''"):
                inheader = False
                indocs = False

            new_content += line
            continue

        if indocs \
           or (line.startswith('"""') or line.startswith("'''") or
               line.startswith('r"""') or line.startswith("r'''")):

            if not indocs:
                # guaranteed to start with """
                if len(line.rstrip()) > 4 \
                  and (line.rstrip().endswith('"""') or line.rstrip().endswith("'''")):
                    # single line doc
                    tmp_line = line.replace('"""', '').replace("'''", '')
                    new_content += f'{sphx_glr_sep}\n'
                    new_content += f'# {tmp_line}'
                else:
                    # must be start of multiline block
                    indocs = True
                    new_content += f'{sphx_glr_sep}\n'
            else:
                # we are already in the docs
                # handle doc end
                if line.rstrip().endswith('"""') or line.rstrip().endswith("'''"):
                    # remove quotes
                    # import ipdb; ipdb.set_trace()
                    indocs = False
                    new_content += '\n'
                else:
                    # import ipdb; ipdb.set_trace()
                    new_content += f'# {line}'
                    # has to be documentation
            continue

        new_content += line

    return new_content


def folder_explicit_order():
    srcdir = os.path.abspath(os.path.dirname(__file__))
    examples_dir = os.path.join(srcdir, '..', 'examples')

    f_example_desc = Path(examples_dir, '_valid_examples.toml')
    if not f_example_desc.exists():
        msg = f'No valid examples description file found in {examples_dir}'
        msg += "(e.g '_valid_examples.toml')"
        abort(msg)

    with open(f_example_desc, 'rb') as fobj:
        try:
            desc_examples = tomllib.load(fobj)
        except Exception as e:
            msg = f'Error Loading examples description file: {e}.\n\n'
            msg += 'Please check the file format.'
            abort(msg)

    if 'main' not in desc_examples.keys():
        msg = 'No main section found in examples description file'
        abort(msg)

    try:
        folder_list = sorted(
            [examplesConfig(folder_name=k.lower(), **v)
             for k, v in desc_examples.items()]
        )
    except Exception as e:
        msg = f'Error parsing examples description file: {e}.\n\n'
        msg += 'Please check the file format.'
        abort(msg)

    return [f.folder_name for f in folder_list if f.enable]


def preprocess_include_directive(input_rst, output_rst):
    """
    Process include directives from input RST, and write the output to a new RST.

    Parameters
    ----------
    input_rst : str
        Path to the input RST file containing the include directive.
    output_rst : str
        Path to the output RST file with the include content expanded.

    """
    with open(input_rst, "r") as infile, open(output_rst, "w") as outfile:
        for line in infile:
            if line.strip().startswith(".. include::"):
                # Extract the included file's path
                included_file_path = line.strip().split(" ")[-1]
                included_file_path = os.path.normpath(
                    os.path.join(os.path.dirname(input_rst), included_file_path)
                )

                # Check if the included file exists and read its content
                if os.path.isfile(included_file_path):
                    with open(included_file_path, "r") as included_file:
                        included_content = included_file.read()
                        # Write the included file's content to the output file
                        outfile.write(included_content)
                else:
                    print(f"Warning: Included file '{included_file_path}' not found.")
            else:
                # Write the line as-is if it's not an include directive
                outfile.write(line)


def prepare_gallery(app=None):
    srcdir = app.srcdir if app else os.path.abspath(pjoin(
        os.path.dirname(__file__), '..'))
    examples_dir = os.path.join(srcdir, 'examples')
    examples_revamp_dir = os.path.join(srcdir, 'examples_revamped')
    os.makedirs(examples_revamp_dir, exist_ok=True)

    f_example_desc = Path(examples_dir, '_valid_examples.toml')
    if not f_example_desc.exists():
        msg = f'No valid examples description file found in {examples_dir}'
        msg += "(e.g '_valid_examples.toml')"
        abort(msg)

    with open(f_example_desc, 'rb') as fobj:
        try:
            desc_examples = tomllib.load(fobj)
        except Exception as e:
            msg = f'Error Loading examples description file: {e}.\n\n'
            msg += 'Please check the file format.'
            abort(msg)

    if 'main' not in desc_examples.keys():
        msg = 'No main section found in examples description file'
        abort(msg)

    try:
        examples_config = sorted(
            [examplesConfig(folder_name=k.lower(), **v)
             for k, v in desc_examples.items()]
        )
    except Exception as e:
        msg = f'Error parsing examples description file: {e}.\n\n'
        msg += 'Please check the file format.'
        abort(msg)

    if examples_config[0].position != 0:
        msg = 'Main section must be first in examples description file with'
        msg += 'position=0'
        abort(msg)
    elif examples_config[0].folder_name != 'main':
        msg = "Main section must be named 'main' in examples description file"
        abort(msg)
    elif examples_config[0].enable is False:
        msg = 'Main section must be enabled in examples description file'
        abort(msg)

    disable_examples_section = []
    included_examples = []
    for example in examples_config:
        if not example.enable:
            disable_examples_section.append(example.folder_name)
            continue

        # Create folder for each example
        if example.position != 0:
            folder = Path(
                examples_revamp_dir, f'{example.folder_name}')
        else:
            folder = Path(examples_revamp_dir)

        if not folder.exists():
            os.makedirs(folder)

        # Create readme file
        if example.readme.startswith('file:'):
            filename = example.readme.split('file:')[1].strip()
            preprocess_include_directive(Path(examples_dir, filename),
                                         Path(folder, "README.rst"))
        else:
            with open(Path(folder, "tmp_readme.rst"), "w", encoding="utf8") as fi:
                fi.write(example.readme)
            preprocess_include_directive(Path(folder, "tmp_readme.rst"),
                                         Path(folder, "README.rst"))
            os.remove(Path(folder, "tmp_readme.rst"))

        # Copy files to folder
        if not example.files:
            continue

        for filename in example.files:
            if not Path(examples_dir, filename).exists():
                msg = f'\tFile {filename} not found in examples folder:  '
                msg += f'{examples_dir}.Please, Add the file or remove it '
                msg += 'from the description file.'
                logger.warning(msg)
                continue

            with open(Path(examples_dir, filename), encoding="utf8") as f:
                xfile = f.readlines()

            new_name = None
            if filename in included_examples:
                # file need to be renamed to make it unique for sphinx-gallery
                occurrences = included_examples.count(fi)
                new_name = f'{filename[:-3]}_{occurrences+1}.py'
            if already_converted(xfile):
                shutil.copy(Path(examples_dir, filename),
                            Path(folder, new_name or filename))
            else:
                with open(Path(folder, new_name or filename), 'w',
                          encoding="utf8") as fi:
                    fi.write(convert_to_sphinx_gallery_format(xfile))
            # Add additional link_names
            with open(Path(folder, new_name or filename), 'r+',
                      encoding="utf8") as fi:
                content = fi.read()
                fi.seek(0, 0)
                link_name = f'\n{sphx_glr_sep}\n'
                link_name += '# .. include:: ../../links_names.inc\n#\n'
                fi.write(content + link_name)

            included_examples.append(filename)

    # Check if all python examples are in the description file
    files_in_config = [fi for ex in examples_config for fi in ex.files]
    all_examples = fnmatch.filter(os.listdir(examples_dir), '*.py')
    for all_ex in all_examples:
        if all_ex in files_in_config:
            continue
        msg = f'File {all_ex} not found in examples '
        msg += f"description file: {f_example_desc}"
        logger.warning(msg)


def setup(app):
    """Install the plugin.
    Parameters
    ----------
    app: Sphinx application context.
    """
    logger.info('Initializing Examples folder revamp plugin...')

    app.connect('builder-inited', prepare_gallery)
    # app.connect('build-finished', summarize_failing_examples)

    metadata = {'parallel_read_safe': True, 'version': app.config.version}

    return metadata


if __name__ == '__main__':
    gallery_name = sys.argv[1]
    outdir = sys.argv[2]

    print(folder_explicit_order())
    prepare_gallery(app=None)



"""Module for utility functions."""

from dipy.utils.convert import expand_range


def handle_vol_idx(vol_idx):
    """Handle user input for volume index.

    Parameters
    ----------
    vol_idx : int, str, list, tuple
        Volume index or range of volume indices.

    Returns
    -------
    vol_idx : list
        List of volume indices.
    """
    if vol_idx is not None:
        if isinstance(vol_idx, str):
            vol_idx = expand_range(vol_idx)
        elif isinstance(vol_idx, int):
            vol_idx = [vol_idx]
        elif isinstance(vol_idx, (list, tuple)):
            vol_idx = [int(idx) for idx in vol_idx]
    return vol_idx



import numpy as np
import pytest

from dipy.stats.resampling import bootstrap, jackknife
from dipy.testing.decorators import set_random_number_generator


@set_random_number_generator(1741332)
def test_bootstrap(rng):
    # Generate a sample data
    x = rng.standard_normal(size=100)

    # Test bootstrap function with default parameters
    bs_pdf, se, ci = bootstrap(x, rng=rng)

    assert len(bs_pdf) == 1000
    assert se > 0

    # Test bootstrap function with custom parameters
    bs_pdf, se, ci = bootstrap(x, statistic=np.mean, B=500, alpha=0.90)

    assert len(bs_pdf) == 500
    assert se > 0


@set_random_number_generator(1741333)
def test_jackknife(rng):
    # Generate a sample data
    pdf = rng.standard_normal(size=100)

    # Test jackknife function with default parameters
    jk_pdf, bias, se = jackknife(pdf, rng=rng)

    assert len(jk_pdf) == 99
    assert bias == pytest.approx(0, abs=1e-1)
    assert se > 0

    # Test jackknife function with custom parameters
    jk_pdf, bias, se = jackknife(pdf, statistic=np.median, M=50)
    assert len(jk_pdf) == 50
    assert se > 0



from collections import defaultdict
import gc


def get_type_refcount(pattern=None):
    """
    Retrieves refcount of types for which their name matches `pattern`.

    Parameters
    ----------
    pattern : str
        Consider only types that have `pattern` in their name.

    Returns
    -------
    dict
        The key is the type name and the value is the refcount.
    """
    gc.collect()

    refcounts_per_type = defaultdict(int)
    for obj in gc.get_objects():
        obj_type_name = type(obj).__name__
        # If `pattern` is not None, keep only matching types.
        if pattern is None or pattern in obj_type_name:
            refcounts_per_type[obj_type_name] += 1

    return refcounts_per_type



""" Utility functions used by the Cross Correlation (CC) metric """

import numpy as np
from dipy.align.fused_types cimport floating
cimport cython
cimport numpy as cnp


cdef inline int _int_max(int a, int b) noexcept nogil:
    r"""
    Returns the maximum of a and b
    """
    return a if a >= b else b


cdef inline int _int_min(int a, int b) noexcept nogil:
    r"""
    Returns the minimum of a and b
    """
    return a if a <= b else b


cdef enum:
    SI = 0
    SI2 = 1
    SJ = 2
    SJ2 = 3
    SIJ = 4
    CNT = 5


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef inline int _wrap(int x, int m) noexcept nogil:
    r""" Auxiliary function to `wrap` an array around its low-end side.
    Negative indices are mapped to last coordinates so that no extra memory
    is required to account for local rectangular windows that exceed the
    array's low-end boundary.

    Parameters
    ----------
    x : int
        the array position to be wrapped
    m : int
        array length
    """
    if x < 0:
        return x + m
    return x


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef inline void _update_factors(double[:, :, :, :] factors,
                                 floating[:, :, :] moving,
                                 floating[:, :, :] static,
                                 cnp.npy_intp ss, cnp.npy_intp rr, cnp.npy_intp cc,
                                 cnp.npy_intp s, cnp.npy_intp r, cnp.npy_intp c, int operation)noexcept nogil:
    r"""Updates the precomputed CC factors of a rectangular window

    Updates the precomputed CC factors of the rectangular window centered
    at (`ss`, `rr`, `cc`) by adding the factors corresponding to voxel
    (`s`, `r`, `c`) of input images `moving` and `static`.

    Parameters
    ----------
    factors : array, shape (S, R, C, 5)
        array containing the current precomputed factors to be updated
    moving : array, shape (S, R, C)
        the moving volume (notice that both images must already be in a common
        reference domain, in particular, they must have the same shape)
    static : array, shape (S, R, C)
        the static volume, which also defines the reference registration domain
    ss : int
        first coordinate of the rectangular window to be updated
    rr : int
        second coordinate of the rectangular window to be updated
    cc : int
        third coordinate of the rectangular window to be updated
    s: int
        first coordinate of the voxel the local window should be updated with
    r: int
        second coordinate of the voxel the local window should be updated with
    c: int
        third coordinate of the voxel the local window should be updated with
    operation : int, either -1, 0 or 1
        indicates whether the factors of voxel (`s`, `r`, `c`) should be
        added to (`operation`=1), subtracted from (`operation`=-1), or set as
        (`operation`=0) the current factors for the rectangular window centered
        at (`ss`, `rr`, `cc`).

    """
    cdef:
        double sval
        double mval
    if s >= moving.shape[0] or r >= moving.shape[1] or c >= moving.shape[2]:
        if operation == 0:
            factors[ss, rr, cc, SI] = 0
            factors[ss, rr, cc, SI2] = 0
            factors[ss, rr, cc, SJ] = 0
            factors[ss, rr, cc, SJ2] = 0
            factors[ss, rr, cc, SIJ] = 0
    else:
        sval = static[s, r, c]
        mval = moving[s, r, c]
        if operation == 0:
            factors[ss, rr, cc, SI] = sval
            factors[ss, rr, cc, SI2] = sval*sval
            factors[ss, rr, cc, SJ] = mval
            factors[ss, rr, cc, SJ2] = mval*mval
            factors[ss, rr, cc, SIJ] = sval*mval
        elif operation == -1:
            factors[ss, rr, cc, SI] -= sval
            factors[ss, rr, cc, SI2] -= sval*sval
            factors[ss, rr, cc, SJ] -= mval
            factors[ss, rr, cc, SJ2] -= mval*mval
            factors[ss, rr, cc, SIJ] -= sval*mval
        elif operation == 1:
            factors[ss, rr, cc, SI] += sval
            factors[ss, rr, cc, SI2] += sval*sval
            factors[ss, rr, cc, SJ] += mval
            factors[ss, rr, cc, SJ2] += mval*mval
            factors[ss, rr, cc, SIJ] += sval*mval


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def precompute_cc_factors_3d(floating[:, :, :] static,
                             floating[:, :, :] moving,
                             cnp.npy_intp radius, num_threads=None):
    """Precomputations to quickly compute the gradient of the CC Metric.

    Pre-computes the separate terms of the cross correlation metric and image
    norms at each voxel considering a neighborhood of the given radius to
    efficiently compute the gradient of the metric with respect to the
    deformation field :footcite:p:`Ocegueda2016`, :footcite:p:`Avants2008`,
    :footcite:p:`Avants2009`.

    Parameters
    ----------
    static : array, shape (S, R, C)
        the static volume, which also defines the reference registration domain
    moving : array, shape (S, R, C)
        the moving volume (notice that both images must already be in a common
        reference domain, i.e. the same S, R, C)
    radius : the radius of the neighborhood (cube of (2 * radius + 1)^3 voxels)

    Returns
    -------
    factors : array, shape (S, R, C, 5)
        the precomputed cross correlation terms::

            - factors[:,:,:,0] : static minus its mean value along the neighborhood
            - factors[:,:,:,1] : moving minus its mean value along the neighborhood
            - factors[:,:,:,2] : sum of the pointwise products of static and moving
              along the neighborhood
            - factors[:,:,:,3] : sum of sq. values of static along the neighborhood
            - factors[:,:,:,4] : sum of sq. values of moving along the neighborhood

    References
    ----------
    .. footbibliography::
    """
    cdef:
        cnp.npy_intp ns = static.shape[0]
        cnp.npy_intp nr = static.shape[1]
        cnp.npy_intp nc = static.shape[2]
        cnp.npy_intp side = 2 * radius + 1
        cnp.npy_intp firstc, lastc, firstr, lastr, firsts, lasts
        cnp.npy_intp s, r, c, it, sides, sider, sidec
        double cnt
        cnp.npy_intp ssss, sss, ss, rr, cc, prev_ss, prev_rr, prev_cc
        double Imean, Jmean, IJprods, Isq, Jsq
        double[:, :, :, :] temp = np.zeros((2, nr, nc, 5), dtype=np.float64)
        floating[:, :, :, :] factors = np.zeros((ns, nr, nc, 5),
                                                dtype=np.asarray(static).dtype)

    with nogil:
        sss = 1
        for s in range(ns+radius):
            ss = _wrap(s - radius, ns)
            sss = 1 - sss
            firsts = _int_max(0, ss - radius)
            lasts = _int_min(ns - 1, ss + radius)
            sides = (lasts - firsts + 1)
            for r in range(nr+radius):
                rr = _wrap(r - radius, nr)
                firstr = _int_max(0, rr - radius)
                lastr = _int_min(nr - 1, rr + radius)
                sider = (lastr - firstr + 1)
                for c in range(nc+radius):
                    cc = _wrap(c - radius, nc)
                    # New corner
                    _update_factors(temp, moving, static,
                                    sss, rr, cc, s, r, c, 0)

                    # Add signed sub-volumes
                    if s > 0:
                        prev_ss = 1 - sss
                        for it in range(5):
                            temp[sss, rr, cc, it] += temp[prev_ss, rr, cc, it]
                        if r > 0:
                            prev_rr = _wrap(rr-1, nr)
                            for it in range(5):
                                temp[sss, rr, cc, it] -= \
                                    temp[prev_ss, prev_rr, cc, it]
                            if c > 0:
                                prev_cc = _wrap(cc-1, nc)
                                for it in range(5):
                                    temp[sss, rr, cc, it] += \
                                        temp[prev_ss, prev_rr, prev_cc, it]
                        if c > 0:
                            prev_cc = _wrap(cc-1, nc)
                            for it in range(5):
                                temp[sss, rr, cc, it] -= \
                                    temp[prev_ss, rr, prev_cc, it]
                    if r > 0:
                        prev_rr = _wrap(rr-1, nr)
                        for it in range(5):
                            temp[sss, rr, cc, it] += \
                                temp[sss, prev_rr, cc, it]
                        if c > 0:
                            prev_cc = _wrap(cc-1, nc)
                            for it in range(5):
                                temp[sss, rr, cc, it] -= \
                                    temp[sss, prev_rr, prev_cc, it]
                    if c > 0:
                        prev_cc = _wrap(cc-1, nc)
                        for it in range(5):
                            temp[sss, rr, cc, it] += temp[sss, rr, prev_cc, it]

                    # Add signed corners
                    if s >= side:
                        _update_factors(temp, moving, static,
                                        sss, rr, cc, s-side, r, c, -1)
                        if r >= side:
                            _update_factors(temp, moving, static,
                                            sss, rr, cc, s-side, r-side, c, 1)
                            if c >= side:
                                _update_factors(temp, moving, static, sss, rr,
                                                cc, s-side, r-side, c-side, -1)
                        if c >= side:
                            _update_factors(temp, moving, static,
                                            sss, rr, cc, s-side, r, c-side, 1)
                    if r >= side:
                        _update_factors(temp, moving, static,
                                        sss, rr, cc, s, r-side, c, -1)
                        if c >= side:
                            _update_factors(temp, moving, static,
                                            sss, rr, cc, s, r-side, c-side, 1)

                    if c >= side:
                        _update_factors(temp, moving, static,
                                        sss, rr, cc, s, r, c-side, -1)
                    # Compute final factors
                    if s >= radius and r >= radius and c >= radius:
                        firstc = _int_max(0, cc - radius)
                        lastc = _int_min(nc - 1, cc + radius)
                        sidec = (lastc - firstc + 1)
                        cnt = sides*sider*sidec
                        Imean = temp[sss, rr, cc, SI] / cnt
                        Jmean = temp[sss, rr, cc, SJ] / cnt
                        IJprods = (temp[sss, rr, cc, SIJ] -
                                   Jmean * temp[sss, rr, cc, SI] -
                                   Imean * temp[sss, rr, cc, SJ] +
                                   cnt * Jmean * Imean)
                        Isq = (temp[sss, rr, cc, SI2] -
                               Imean * temp[sss, rr, cc, SI] -
                               Imean * temp[sss, rr, cc, SI] +
                               cnt * Imean * Imean)
                        Jsq = (temp[sss, rr, cc, SJ2] -
                               Jmean * temp[sss, rr, cc, SJ] -
                               Jmean * temp[sss, rr, cc, SJ] +
                               cnt * Jmean * Jmean)
                        factors[ss, rr, cc, 0] = static[ss, rr, cc] - Imean
                        factors[ss, rr, cc, 1] = moving[ss, rr, cc] - Jmean
                        factors[ss, rr, cc, 2] = IJprods
                        factors[ss, rr, cc, 3] = Isq
                        factors[ss, rr, cc, 4] = Jsq
    return factors


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def precompute_cc_factors_3d_test(floating[:, :, :] static,
                                  floating[:, :, :] moving, int radius):
    """Precomputations to quickly compute the gradient of the CC Metric.

    This version of precompute_cc_factors_3d is for testing purposes, it
    directly computes the local cross-correlation factors without any
    optimization, so it is less error-prone than the accelerated version.
    """
    cdef:
        cnp.npy_intp ns = static.shape[0]
        cnp.npy_intp nr = static.shape[1]
        cnp.npy_intp nc = static.shape[2]
        cnp.npy_intp s, r, c, k, i, j, t
        cnp.npy_intp firstc, lastc, firstr, lastr, firsts, lasts
        double Imean, Jmean
        floating[:, :, :, :] factors = np.zeros((ns, nr, nc, 5),
                                                dtype=np.asarray(static).dtype)
        double[:] sums = np.zeros((6,), dtype=np.float64)

    with nogil:
        for s in range(ns):
            firsts = _int_max(0, s - radius)
            lasts = _int_min(ns - 1, s + radius)
            for r in range(nr):
                firstr = _int_max(0, r - radius)
                lastr = _int_min(nr - 1, r + radius)
                for c in range(nc):
                    firstc = _int_max(0, c - radius)
                    lastc = _int_min(nc - 1, c + radius)
                    for t in range(6):
                        sums[t] = 0
                    for k in range(firsts, 1 + lasts):
                        for i in range(firstr, 1 + lastr):
                            for j in range(firstc, 1 + lastc):
                                sums[SI] += static[k, i, j]
                                sums[SI2] += static[k, i, j]**2
                                sums[SJ] += moving[k, i, j]
                                sums[SJ2] += moving[k, i, j]**2
                                sums[SIJ] += static[k, i, j]*moving[k, i, j]
                                sums[CNT] += 1
                    Imean = sums[SI] / sums[CNT]
                    Jmean = sums[SJ] / sums[CNT]
                    factors[s, r, c, 0] = static[s, r, c] - Imean
                    factors[s, r, c, 1] = moving[s, r, c] - Jmean
                    factors[s, r, c, 2] = (sums[SIJ] - Jmean * sums[SI] -
                                           Imean * sums[SJ] +
                                           sums[CNT] * Jmean * Imean)
                    factors[s, r, c, 3] = (sums[SI2] - Imean * sums[SI] -
                                           Imean * sums[SI] +
                                           sums[CNT] * Imean * Imean)
                    factors[s, r, c, 4] = (sums[SJ2] - Jmean * sums[SJ] -
                                           Jmean * sums[SJ] +
                                           sums[CNT] * Jmean * Jmean)
    return np.asarray(factors)


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def compute_cc_forward_step_3d(floating[:, :, :, :] grad_static,
                               floating[:, :, :, :] factors,
                               cnp.npy_intp radius):
    """Gradient of the CC Metric w.r.t. the forward transformation.

    Computes the gradient of the Cross Correlation metric for symmetric
    registration (SyN) :footcite:p:`Avants2008` w.r.t. the displacement
    associated to the moving volume ('forward' step) as in
    :footcite:t:`Avants2009`.

    Parameters
    ----------
    grad_static : array, shape (S, R, C, 3)
        the gradient of the static volume
    factors : array, shape (S, R, C, 5)
        the precomputed cross correlation terms obtained via
        precompute_cc_factors_3d
    radius : int
        the radius of the neighborhood used for the CC metric when
        computing the factors. The returned vector field will be
        zero along a boundary of width radius voxels.

    Returns
    -------
    out : array, shape (S, R, C, 3)
        the gradient of the cross correlation metric with respect to the
        displacement associated to the moving volume
    energy : the cross correlation energy (data term) at this iteration

    References
    ----------
    .. footbibliography::
    """
    cdef:
        cnp.npy_intp ns = grad_static.shape[0]
        cnp.npy_intp nr = grad_static.shape[1]
        cnp.npy_intp nc = grad_static.shape[2]
        double energy = 0
        cnp.npy_intp s, r, c
        double Ii, Ji, sfm, sff, smm, localCorrelation, temp
        floating[:, :, :, :] out =\
            np.zeros((ns, nr, nc, 3), dtype=np.asarray(grad_static).dtype)
    with nogil:
        for s in range(radius, ns-radius):
            for r in range(radius, nr-radius):
                for c in range(radius, nc-radius):
                    Ii = factors[s, r, c, 0]
                    Ji = factors[s, r, c, 1]
                    sfm = factors[s, r, c, 2]
                    sff = factors[s, r, c, 3]
                    smm = factors[s, r, c, 4]
                    if sff == 0.0 or smm == 0.0:
                        continue
                    localCorrelation = 0
                    if sff * smm > 1e-5:
                        localCorrelation = sfm * sfm / (sff * smm)
                    if localCorrelation < 1:  # avoid bad values...
                        energy -= localCorrelation
                    temp = 2.0 * sfm / (sff * smm) * (Ji - sfm / sff * Ii)
                    out[s, r, c, 0] -= temp * grad_static[s, r, c, 0]
                    out[s, r, c, 1] -= temp * grad_static[s, r, c, 1]
                    out[s, r, c, 2] -= temp * grad_static[s, r, c, 2]
    return np.asarray(out), energy


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def compute_cc_backward_step_3d(floating[:, :, :, :] grad_moving,
                                floating[:, :, :, :] factors,
                                cnp.npy_intp radius):
    """Gradient of the CC Metric w.r.t. the backward transformation.

    Computes the gradient of the Cross Correlation metric for symmetric
    registration (SyN) :footcite:p:`Avants2008`. w.r.t. the displacement
    associated to the static volume ('backward' step) as in
    :footcite:t:`Avants2009`.

    Parameters
    ----------
    grad_moving : array, shape (S, R, C, 3)
        the gradient of the moving volume
    factors : array, shape (S, R, C, 5)
        the precomputed cross correlation terms obtained via
        precompute_cc_factors_3d
    radius : int
        the radius of the neighborhood used for the CC metric when
        computing the factors. The returned vector field will be
        zero along a boundary of width radius voxels.

    Returns
    -------
    out : array, shape (S, R, C, 3)
        the gradient of the cross correlation metric with respect to the
        displacement associated to the static volume
    energy : the cross correlation energy (data term) at this iteration

    References
    ----------
    .. footbibliography::
    """
    ftype = np.asarray(grad_moving).dtype
    cdef:
        cnp.npy_intp ns = grad_moving.shape[0]
        cnp.npy_intp nr = grad_moving.shape[1]
        cnp.npy_intp nc = grad_moving.shape[2]
        cnp.npy_intp s, r, c
        double energy = 0
        double Ii, Ji, sfm, sff, smm, localCorrelation, temp
        floating[:, :, :, :] out = np.zeros((ns, nr, nc, 3), dtype=ftype)

    with nogil:

        for s in range(radius, ns-radius):
            for r in range(radius, nr-radius):
                for c in range(radius, nc-radius):
                    Ii = factors[s, r, c, 0]
                    Ji = factors[s, r, c, 1]
                    sfm = factors[s, r, c, 2]
                    sff = factors[s, r, c, 3]
                    smm = factors[s, r, c, 4]
                    if sff == 0.0 or smm == 0.0:
                        continue
                    localCorrelation = 0
                    if sff * smm > 1e-5:
                        localCorrelation = sfm * sfm / (sff * smm)
                    if localCorrelation < 1:  # avoid bad values...
                        energy -= localCorrelation
                    temp = 2.0 * sfm / (sff * smm) * (Ii - sfm / smm * Ji)
                    out[s, r, c, 0] -= temp * grad_moving[s, r, c, 0]
                    out[s, r, c, 1] -= temp * grad_moving[s, r, c, 1]
                    out[s, r, c, 2] -= temp * grad_moving[s, r, c, 2]
    return np.asarray(out), energy


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def precompute_cc_factors_2d(floating[:, :] static, floating[:, :] moving,
                             cnp.npy_intp radius):
    """Precomputations to quickly compute the gradient of the CC Metric.

    Pre-computes the separate terms of the cross correlation metric
    :footcite:p:`Avants2008` and image norms at each voxel considering a
    neighborhood of the given radius to efficiently compute the gradient of the
    metric with respect to the deformation field :footcite:p:`Avants2009`.

    Parameters
    ----------
    static : array, shape (R, C)
        the static volume, which also defines the reference registration domain
    moving : array, shape (R, C)
        the moving volume (notice that both images must already be in a common
        reference domain, i.e. the same R, C)
    radius : the radius of the neighborhood(square of (2*radius + 1)^2 voxels)

    Returns
    -------
    factors : array, shape (R, C, 5)
        the precomputed cross correlation terms::

            - factors[:,:,0] : static minus its mean value along the neighborhood
            - factors[:,:,1] : moving minus its mean value along the neighborhood
            - factors[:,:,2] : sum of the pointwise products of static and moving
              along the neighborhood
            - factors[:,:,3] : sum of sq. values of static along the neighborhood
            - factors[:,:,4] : sum of sq. values of moving along the neighborhood

    References
    ----------
    .. footbibliography::
    """
    ftype = np.asarray(static).dtype
    cdef:
        cnp.npy_intp side = 2 * radius + 1
        cnp.npy_intp nr = static.shape[0]
        cnp.npy_intp nc = static.shape[1]
        cnp.npy_intp r, c, i, j, t, q, qq, firstc, lastc
        double Imean, Jmean
        floating[:, :, :] factors = np.zeros((nr, nc, 5), dtype=ftype)
        double[:, :] lines = np.zeros((6, side), dtype=np.float64)
        double[:] sums = np.zeros((6,), dtype=np.float64)

    with nogil:

        for c in range(nc):
            firstc = _int_max(0, c - radius)
            lastc = _int_min(nc - 1, c + radius)
            # compute factors for row [:,c]
            for t in range(6):
                for q in range(side):
                    lines[t, q] = 0
            # Compute all rows and set the sums on the fly
            # compute row [i, j = {c-radius, c + radius}]
            for i in range(nr):
                q = i % side
                for t in range(6):
                    lines[t, q] = 0
                for j in range(firstc, lastc + 1):
                    lines[SI, q] += static[i, j]
                    lines[SI2, q] += static[i, j] * static[i, j]
                    lines[SJ, q] += moving[i, j]
                    lines[SJ2, q] += moving[i, j] * moving[i, j]
                    lines[SIJ, q] += static[i, j] * moving[i, j]
                    lines[CNT, q] += 1

                for t in range(6):
                    sums[t] = 0
                    for qq in range(side):
                        sums[t] += lines[t, qq]
                if i >= radius:
                    # r is the pixel that is affected by the cube with slices
                    # [r - radius.. r + radius, :]
                    r = i - radius
                    Imean = sums[SI] / sums[CNT]
                    Jmean = sums[SJ] / sums[CNT]
                    factors[r, c, 0] = static[r, c] - Imean
                    factors[r, c, 1] = moving[r, c] - Jmean
                    factors[r, c, 2] = (sums[SIJ] - Jmean * sums[SI] -
                                        Imean * sums[SJ] +
                                        sums[CNT] * Jmean * Imean)
                    factors[r, c, 3] = (sums[SI2] - Imean * sums[SI] -
                                        Imean * sums[SI] +
                                        sums[CNT] * Imean * Imean)
                    factors[r, c, 4] = (sums[SJ2] - Jmean * sums[SJ] -
                                        Jmean * sums[SJ] +
                                        sums[CNT] * Jmean * Jmean)
            # Finally set the values at the end of the line
            for r in range(nr - radius, nr):
                # this would be the last slice to be processed for pixel
                # [r, c], if it existed
                i = r + radius
                q = i % side
                for t in range(6):
                    sums[t] -= lines[t, q]
                Imean = sums[SI] / sums[CNT]
                Jmean = sums[SJ] / sums[CNT]
                factors[r, c, 0] = static[r, c] - Imean
                factors[r, c, 1] = moving[r, c] - Jmean
                factors[r, c, 2] = (sums[SIJ] - Jmean * sums[SI] -
                                    Imean * sums[SJ] +
                                    sums[CNT] * Jmean * Imean)
                factors[r, c, 3] = (sums[SI2] - Imean * sums[SI] -
                                    Imean * sums[SI] +
                                    sums[CNT] * Imean * Imean)
                factors[r, c, 4] = (sums[SJ2] - Jmean * sums[SJ] -
                                    Jmean * sums[SJ] +
                                    sums[CNT] * Jmean * Jmean)
    return np.asarray(factors)


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def precompute_cc_factors_2d_test(floating[:, :] static, floating[:, :] moving,
                                  cnp.npy_intp radius):
    """Precomputations to quickly compute the gradient of the CC Metric.

    This version of precompute_cc_factors_2d is for testing purposes, it
    directly computes the local cross-correlation without any optimization.
    """
    ftype = np.asarray(static).dtype
    cdef:
        cnp.npy_intp nr = static.shape[0]
        cnp.npy_intp nc = static.shape[1]
        cnp.npy_intp r, c, i, j, t, firstr, lastr, firstc, lastc
        double Imean, Jmean
        floating[:, :, :] factors = np.zeros((nr, nc, 5), dtype=ftype)
        double[:] sums = np.zeros((6,), dtype=np.float64)

    with nogil:

        for r in range(nr):
            firstr = _int_max(0, r - radius)
            lastr = _int_min(nr - 1, r + radius)
            for c in range(nc):
                firstc = _int_max(0, c - radius)
                lastc = _int_min(nc - 1, c + radius)
                for t in range(6):
                    sums[t] = 0
                for i in range(firstr, 1 + lastr):
                    for j in range(firstc, 1+lastc):
                        sums[SI] += static[i, j]
                        sums[SI2] += static[i, j]**2
                        sums[SJ] += moving[i, j]
                        sums[SJ2] += moving[i, j]**2
                        sums[SIJ] += static[i, j]*moving[i, j]
                        sums[CNT] += 1
                Imean = sums[SI] / sums[CNT]
                Jmean = sums[SJ] / sums[CNT]
                factors[r, c, 0] = static[r, c] - Imean
                factors[r, c, 1] = moving[r, c] - Jmean
                factors[r, c, 2] = (sums[SIJ] - Jmean * sums[SI] -
                                    Imean * sums[SJ] +
                                    sums[CNT] * Jmean * Imean)
                factors[r, c, 3] = (sums[SI2] - Imean * sums[SI] -
                                    Imean * sums[SI] +
                                    sums[CNT] * Imean * Imean)
                factors[r, c, 4] = (sums[SJ2] - Jmean * sums[SJ] -
                                    Jmean * sums[SJ] +
                                    sums[CNT] * Jmean * Jmean)
    return np.asarray(factors)


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def compute_cc_forward_step_2d(floating[:, :, :] grad_static,
                               floating[:, :, :] factors,
                               cnp.npy_intp radius):
    """Gradient of the CC Metric w.r.t. the forward transformation.

    Computes the gradient of the Cross Correlation metric for symmetric
    registration (SyN) :footcite:p:`Avants2008` w.r.t. the displacement
    associated to the moving image ('backward' step) as in
    :footcite:t:`Avants2009`.

    Parameters
    ----------
    grad_static : array, shape (R, C, 2)
        the gradient of the static image
    factors : array, shape (R, C, 5)
        the precomputed cross correlation terms obtained via
        precompute_cc_factors_2d

    Returns
    -------
    out : array, shape (R, C, 2)
        the gradient of the cross correlation metric with respect to the
        displacement associated to the moving image
    energy : the cross correlation energy (data term) at this iteration

    Notes
    -----
    Currently, the gradient of the static image is not being used, but some
    authors suggest that symmetrizing the gradient by including both, the
    moving and static gradients may improve the registration quality. We are
    leaving this parameter as a placeholder for future investigation

    References
    ----------
    .. footbibliography::
    """
    cdef:
        cnp.npy_intp nr = grad_static.shape[0]
        cnp.npy_intp nc = grad_static.shape[1]
        double energy = 0
        cnp.npy_intp r, c
        double Ii, Ji, sfm, sff, smm, localCorrelation, temp
        floating[:, :, :] out = np.zeros((nr, nc, 2),
                                         dtype=np.asarray(grad_static).dtype)
    with nogil:

        for r in range(radius, nr-radius):
            for c in range(radius, nc-radius):
                Ii = factors[r, c, 0]
                Ji = factors[r, c, 1]
                sfm = factors[r, c, 2]
                sff = factors[r, c, 3]
                smm = factors[r, c, 4]
                if sff == 0.0 or smm == 0.0:
                    continue
                localCorrelation = 0
                if sff * smm > 1e-5:
                    localCorrelation = sfm * sfm / (sff * smm)
                if localCorrelation < 1:  # avoid bad values...
                    energy -= localCorrelation
                temp = 2.0 * sfm / (sff * smm) * (Ji - sfm / sff * Ii)
                out[r, c, 0] -= temp * grad_static[r, c, 0]
                out[r, c, 1] -= temp * grad_static[r, c, 1]
    return np.asarray(out), energy


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def compute_cc_backward_step_2d(floating[:, :, :] grad_moving,
                                floating[:, :, :] factors,
                                cnp.npy_intp radius):
    """Gradient of the CC Metric w.r.t. the backward transformation.

    Computes the gradient of the Cross Correlation metric for symmetric
    registration (SyN) :footcite:p:`Avants2008` w.r.t. the displacement
    associated to the static image ('forward' step) as in
    :footcite:t:`Avants2009`.

    Parameters
    ----------
    grad_moving : array, shape (R, C, 2)
        the gradient of the moving image
    factors : array, shape (R, C, 5)
        the precomputed cross correlation terms obtained via
        precompute_cc_factors_2d

    Returns
    -------
    out : array, shape (R, C, 2)
        the gradient of the cross correlation metric with respect to the
        displacement associated to the static image
    energy : the cross correlation energy (data term) at this iteration

    References
    ----------
    .. footbibliography::
    """
    ftype = np.asarray(grad_moving).dtype
    cdef:
        cnp.npy_intp nr = grad_moving.shape[0]
        cnp.npy_intp nc = grad_moving.shape[1]
        cnp.npy_intp r, c
        double energy = 0
        double Ii, Ji, sfm, sff, smm, localCorrelation, temp
        floating[:, :, :] out = np.zeros((nr, nc, 2), dtype=ftype)

    with nogil:

        for r in range(radius, nr-radius):
            for c in range(radius, nc-radius):
                Ii = factors[r, c, 0]
                Ji = factors[r, c, 1]
                sfm = factors[r, c, 2]
                sff = factors[r, c, 3]
                smm = factors[r, c, 4]
                if sff == 0.0 or smm == 0.0:
                    continue
                localCorrelation = 0
                if sff * smm > 1e-5:
                    localCorrelation = sfm * sfm / (sff * smm)
                if localCorrelation < 1:  # avoid bad values...
                    energy -= localCorrelation
                temp = 2.0 * sfm / (sff * smm) * (Ii - sfm / smm * Ji)
                out[r, c, 0] -= temp * grad_moving[r, c, 0]
                out[r, c, 1] -= temp * grad_moving[r, c, 1]
    return np.asarray(out), energy



"""
===============================
Calculate DSI-based scalar maps
===============================

We show how to calculate two DSI-based scalar maps: return to origin
probability (RTOP) :footcite:p:`Descoteaux2011` and mean square displacement
(MSD) :footcite:p:`Wu2007`, :footcite:p:`Wu2008` on your dataset.

First import the necessary modules:
"""

import matplotlib.pyplot as plt
import numpy as np

from dipy.core.gradients import gradient_table
from dipy.data import get_fnames
from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti
from dipy.reconst.dsi import DiffusionSpectrumModel

###############################################################################
# Download and get the data filenames for this tutorial.

fraw, fbval, fbvec = get_fnames(name="taiwan_ntu_dsi")

###############################################################################
# img contains a nibabel Nifti1Image object (data) and gtab contains a
# GradientTable object (gradient information e.g. b-values). For example to
# read the b-values it is possible to write print(gtab.bvals).
#
# Load the raw diffusion data and the affine.

data, affine = load_nifti(fraw)
bvals, bvecs = read_bvals_bvecs(fbval, fbvec)
bvecs[1:] = bvecs[1:] / np.sqrt(np.sum(bvecs[1:] * bvecs[1:], axis=1))[:, None]
gtab = gradient_table(bvals, bvecs=bvecs)

print(f"data.shape {data.shape}")

###############################################################################
# Instantiate the Model and apply it to the data.

dsmodel = DiffusionSpectrumModel(gtab, qgrid_size=35, filter_width=18.5)

###############################################################################
# Let's just use one slice only from the data.

dataslice = data[30:70, 20:80, data.shape[2] // 2]

###############################################################################
# Normalize the signal by the b0

dataslice = dataslice / (dataslice[..., 0, None]).astype(float)

###############################################################################
# Calculate the return to origin probability on the signal
# that corresponds to the integral of the signal.

print("Calculating... rtop_signal")
rtop_signal = dsmodel.fit(dataslice).rtop_signal()

###############################################################################
# Now we calculate the return to origin probability on the propagator, that
# corresponds to its central value. By default the propagator is divided by
# its sum in order to obtain a properly normalized pdf, however this
# normalization changes the values of RTOP, therefore in order to compare it
# with the RTOP previously calculated on the signal we turn the normalized
# parameter to false.

print("Calculating... rtop_pdf")
rtop_pdf = dsmodel.fit(dataslice).rtop_pdf(normalized=False)

###############################################################################
# In theory, these two measures must be equal,
# to show that we calculate the mean square error on this two measures.

mse = np.sum((rtop_signal - rtop_pdf) ** 2) / rtop_signal.size
print(f"mse = {mse:f}")

###############################################################################
# Leaving the normalized parameter to the default changes the values of the
# RTOP but not the contrast between the voxels.

print("Calculating... rtop_pdf_norm")
rtop_pdf_norm = dsmodel.fit(dataslice).rtop_pdf()

###############################################################################
# Let's calculate the mean square displacement on the normalized propagator.

print("Calculating... msd_norm")
msd_norm = dsmodel.fit(dataslice).msd_discrete()

###############################################################################
# Turning the normalized parameter to false makes it possible to calculate
# the mean square displacement on the propagator without normalization.

print("Calculating... msd")
msd = dsmodel.fit(dataslice).msd_discrete(normalized=False)

###############################################################################
# Show the RTOP images and save them in rtop.png.

fig = plt.figure(figsize=(6, 6))
ax1 = fig.add_subplot(2, 2, 1, title="rtop_signal")
ax1.set_axis_off()
ind = ax1.imshow(rtop_signal.T, interpolation="nearest", origin="lower")
plt.colorbar(ind)
ax2 = fig.add_subplot(2, 2, 2, title="rtop_pdf_norm")
ax2.set_axis_off()
ind = ax2.imshow(rtop_pdf_norm.T, interpolation="nearest", origin="lower")
plt.colorbar(ind)
ax3 = fig.add_subplot(2, 2, 3, title="rtop_pdf")
ax3.set_axis_off()
ind = ax3.imshow(rtop_pdf.T, interpolation="nearest", origin="lower")
plt.colorbar(ind)
plt.savefig("rtop.png")

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Return to origin probability.
#
#
# Show the MSD images and save them in msd.png.

fig = plt.figure(figsize=(7, 3))
ax1 = fig.add_subplot(1, 2, 1, title="msd_norm")
ax1.set_axis_off()
ind = ax1.imshow(msd_norm.T, interpolation="nearest", origin="lower")
plt.colorbar(ind)
ax2 = fig.add_subplot(1, 2, 2, title="msd")
ax2.set_axis_off()
ind = ax2.imshow(msd.T, interpolation="nearest", origin="lower")
plt.colorbar(ind)
plt.savefig("msd.png")

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Mean square displacement.
#
#
# References
# ----------
#
# .. footbibliography::
#



#!/usr/bin/env python3
# Script to upload docs to gh-pages branch of dipy_web that will be
# automatically detected by the dipy website.
import os
from os import chdir as cd
import re
from subprocess import check_call
import sys


def sh(cmd):
    """Execute command in a subshell, return status code."""
    print("--------------------------------------------------")
    print(f"Executing: {cmd}")
    print("--------------------------------------------------")
    return check_call(cmd, shell=True)


# paths
docs_repo_path = "_build/docs_repo"
docs_repo_url = "https://github.com/dipy/dipy_web.git"

if __name__ == '__main__':
    # get current directory
    startdir = os.getcwd()

    # find the source version
    info_file = '../dipy/info.py'
    info_lines = open(info_file).readlines()
    source_version = '.'.join(
        [v.split('=')[1].strip(" '\n.")
         for v in info_lines if re.match(
            '^_version_(major|minor|micro|extra)',
            v)])
    print("Source version: ", source_version)

    # check for dev tag
    if source_version.split(".")[-1] == "dev":
        dev = True
        print("Development Version detected")
    else:
        dev = False

    # pull current docs_repo
    if not os.path.exists(docs_repo_path):
        print("docs_repo not found, pulling from git..")
        sh(f"git clone {docs_repo_url} {docs_repo_path}")
    cd(docs_repo_path)
    print(f"Moved to {os.getcwd()}")
    try:
        sh("git checkout gh-pages")
    except:
        while 1:
            print("\nLooks like gh-pages branch does not exist!")
            print("Do you want to create a new one? (y/n)")
            choice = str(input()).lower()
            if choice == 'y':
                sh("git checkout -b gh-pages")
                sh("rm -rf *")
                sh("git add .")
                sh("git commit -m 'cleaning gh-pages branch'")
                sh("git push origin gh-pages")
                break
            if choice == 'n':
                print("Please manually create a new gh-pages branch and try again.")
                sys.exit(0)
            else:
                print("Please enter valid choice ..")
    sh("git pull origin gh-pages")

    # check if docs for current version exists
    if os.path.exists(source_version) and not dev:
        print("docs for current version already exists")
    else:
        if dev:
            print("Re-building docs for development version")
        else:
            print("Building docs for a release")
        # build docs and copy to docs_repo
        cd(startdir)
        # remove old html and doctree files
        try:
            sh("rm -rf _build/json _build/doctrees")
        except:
            pass
        # generate new doc and copy to docs_repo
        sh("make api")
        sh("make rstexamples")
        sh("make json")
        sh(f"cp -r _build/json {docs_repo_path}/")
        cd(docs_repo_path)
        if dev:
            try:
                sh(f"rm -r {source_version}")
            except:
                pass
        sh(f"mv json {source_version}")
        sh("git add .")
        sh(f"git commit -m \"Add docs for {source_version}\"")
        sh("git push origin gh-pages")



import logging
from os.path import join as pjoin
from tempfile import TemporaryDirectory
import warnings

import numpy as np
import numpy.testing as npt

from dipy.core.gradients import generate_bvecs
from dipy.data import get_fnames
from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti, load_nifti_data, save_nifti
from dipy.io.peaks import load_pam
from dipy.reconst.shm import descoteaux07_legacy_msg, sph_harm_ind_list
from dipy.workflows.reconst import ReconstCSDFlow, ReconstQBallBaseFlow, ReconstSDTFlow

logging.getLogger().setLevel(logging.INFO)


def test_reconst_csa():
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        reconst_flow_core(ReconstQBallBaseFlow)


def test_reconst_opdt():
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        reconst_flow_core(ReconstQBallBaseFlow, method="opdt")


def test_reconst_qball():
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        reconst_flow_core(ReconstQBallBaseFlow, method="qball")


def test_reconst_csd():
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        reconst_flow_core(ReconstCSDFlow)


def test_reconst_sdt():
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        reconst_flow_core(ReconstSDTFlow)


def reconst_flow_core(flow, **kwargs):
    with TemporaryDirectory() as out_dir:
        data_path, bval_path, bvec_path = get_fnames(name="small_64D")
        volume, affine = load_nifti(data_path)
        mask = np.ones_like(volume[:, :, :, 0])
        mask_path = pjoin(out_dir, "tmp_mask.nii.gz")
        save_nifti(mask_path, mask.astype(np.uint8), affine)

        reconst_flow = flow()
        for sh_order in [4, 6, 8]:
            reconst_flow.run(
                data_path,
                bval_path,
                bvec_path,
                mask_path,
                sh_order_max=sh_order,
                out_dir=out_dir,
                extract_pam_values=True,
                **kwargs,
            )

            gfa_path = reconst_flow.last_generated_outputs["out_gfa"]
            gfa_data = load_nifti_data(gfa_path)
            npt.assert_equal(gfa_data.shape, volume.shape[:-1])

            peaks_dir_path = reconst_flow.last_generated_outputs["out_peaks_dir"]
            peaks_dir_data = load_nifti_data(peaks_dir_path)
            npt.assert_equal(peaks_dir_data.shape[-1], 15)
            npt.assert_equal(peaks_dir_data.shape[:-1], volume.shape[:-1])

            peaks_idx_path = reconst_flow.last_generated_outputs["out_peaks_indices"]
            peaks_idx_data = load_nifti_data(peaks_idx_path)
            npt.assert_equal(peaks_idx_data.shape[-1], 5)
            npt.assert_equal(peaks_idx_data.shape[:-1], volume.shape[:-1])

            peaks_vals_path = reconst_flow.last_generated_outputs["out_peaks_values"]
            peaks_vals_data = load_nifti_data(peaks_vals_path)
            npt.assert_equal(peaks_vals_data.shape[-1], 5)
            npt.assert_equal(peaks_vals_data.shape[:-1], volume.shape[:-1])

            shm_path = reconst_flow.last_generated_outputs["out_shm"]
            shm_data = load_nifti_data(shm_path)
            # Test that the number of coefficients is what you would expect
            # given the order of the sh basis:
            npt.assert_equal(
                shm_data.shape[-1], sph_harm_ind_list(sh_order)[0].shape[0]
            )
            npt.assert_equal(shm_data.shape[:-1], volume.shape[:-1])

            pam = load_pam(reconst_flow.last_generated_outputs["out_pam"])
            npt.assert_allclose(
                pam.peak_dirs.reshape(peaks_dir_data.shape), peaks_dir_data
            )
            npt.assert_allclose(pam.peak_values, peaks_vals_data)
            npt.assert_allclose(pam.peak_indices, peaks_idx_data)
            npt.assert_allclose(pam.shm_coeff, shm_data)
            npt.assert_allclose(pam.gfa, gfa_data)

            bvals, bvecs = read_bvals_bvecs(bval_path, bvec_path)
            bvals[0] = 5.0
            bvecs = generate_bvecs(len(bvals))

            tmp_bval_path = pjoin(out_dir, "tmp.bval")
            tmp_bvec_path = pjoin(out_dir, "tmp.bvec")
            np.savetxt(tmp_bval_path, bvals)
            np.savetxt(tmp_bvec_path, bvecs.T)
            reconst_flow._force_overwrite = True

            if flow.get_short_name() == "csd":
                reconst_flow = flow()
                reconst_flow._force_overwrite = True
                reconst_flow.run(
                    data_path,
                    bval_path,
                    bvec_path,
                    mask_path,
                    out_dir=out_dir,
                    frf=[15, 5, 5],
                    **kwargs,
                )
                reconst_flow = flow()
                reconst_flow._force_overwrite = True
                reconst_flow.run(
                    data_path,
                    bval_path,
                    bvec_path,
                    mask_path,
                    out_dir=out_dir,
                    frf="15, 5, 5",
                    **kwargs,
                )
                reconst_flow = flow()
                reconst_flow._force_overwrite = True
                reconst_flow.run(
                    data_path,
                    bval_path,
                    bvec_path,
                    mask_path,
                    out_dir=out_dir,
                    frf=None,
                    **kwargs,
                )
                reconst_flow2 = flow()
                reconst_flow2._force_overwrite = True
                reconst_flow2.run(
                    data_path,
                    bval_path,
                    bvec_path,
                    mask_path,
                    out_dir=out_dir,
                    frf=None,
                    roi_center=[5, 5, 5],
                    **kwargs,
                )
            else:
                with npt.assert_raises(BaseException):
                    npt.assert_warns(
                        UserWarning,
                        reconst_flow.run,
                        data_path,
                        tmp_bval_path,
                        tmp_bvec_path,
                        mask_path,
                        out_dir=out_dir,
                        extract_pam_values=True,
                        **kwargs,
                    )

            # test parallel implementation
            # Avoid SDT for now, as it is quite slow, something to introspect
            if flow.get_short_name() != "sdt":
                reconst_flow = flow()
                reconst_flow._force_overwrite = True
                reconst_flow.run(
                    data_path,
                    bval_path,
                    bvec_path,
                    mask_path,
                    out_dir=out_dir,
                    parallel=True,
                    num_processes=2,
                    **kwargs,
                )
